#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import cgi
import logging
from copy import deepcopy
from typing import Dict, List, Optional, Union

import pandas
from ads.common import utils
from ads.config import COMPARTMENT_OCID, PROJECT_OCID
from ads.feature_engineering.schema import Schema
from ads.jobs.builders.base import Builder
from ads.model.model_metadata import (
    ModelCustomMetadata,
    ModelProvenanceMetadata,
    ModelTaxonomyMetadata,
)
from ads.model.service.oci_datascience_model import (
    ModelProvenanceNotFoundError,
    OCIDataScienceModel,
)
from ads.model.artifact_downloader import (
    LargeArtifactDownloader,
    SmallArtifactDownloader,
)
from ads.model.artifact_uploader import LargeArtifactUploader, SmallArtifactUploader

logger = logging.getLogger(__name__)


_MAX_ARTIFACT_SIZE_IN_BYTES = 2147483648  # 2GB


class ModelArtifactSizeError(Exception):
    def __init__(self, max_artifact_size: str):
        super().__init__(
            f"The model artifacts size is greater than `{max_artifact_size}`. "
            "The `bucket_uri` needs to be specified to "
            "copy artifacts to the object storage bucket. "
            "Example: `bucket_uri=oci://<bucket_name>@<namespace>/prefix/`"
        )


class DataScienceModel(Builder):
    """Represents a Data Science Model.

    Attributes
    ----------
    id: str
        Model ID.
    project_id: str
        Project OCID.
    compartment_id: str
        Compartment OCID.
    name: str
        Model name.
    description: str
        Model description.
    freeform_tags: Dict[str, str]
        Model freeform tags.
    defined_tags: Dict[str, Dict[str, object]]
        Model defined tags.
    input_schema: ads.feature_engineering.Schema
        Model input schema.
    output_schema: ads.feature_engineering.Schema, Dict
        Model output schema.
    defined_metadata_list: ModelTaxonomyMetadata
        Model defined metadata.
    custom_metadata_list: ModelCustomMetadata
        Model custom metadata.
    provenance_metadata: ModelProvenanceMetadata
        Model provenance metadata.
    artifact: str
        The artifact location. Can be either path to folder with artifacts or
        path to zip archive.
    status: Union[str, None]
        Model status.
    model_version_set_id: str
        Model version set ID
    version_label: str
        Model version label

    Methods
    -------
    create(self, **kwargs) -> "DataScienceModel"
        Creates model.
    delete(self, delete_associated_model_deployment: Optional[bool] = False) -> "DataScienceModel":
        Removes model.
    to_dict(self) -> dict
        Serializes model to a dictionary.
    from_id(cls, id: str) -> "DataScienceModel"
        Gets an existing model by OCID.
    from_dict(cls, config: dict) -> "DataScienceModel"
        Loads model instance from a dictionary of configurations.
    upload_artifact(self, ...) -> None
        Uploads model artifacts to the model catalog.
    download_artifact(self, ...) -> None
        Downloads model artifacts from the model catalog.
    update(self, **kwargs) -> "DataScienceModel"
        Updates datascience model in model catalog.
    list(cls, compartment_id: str = None, **kwargs) -> List["DataScienceModel"]
        Lists datascience models in a given compartment.
    sync(self):
        Sync up a datascience model with OCI datascience model.
    with_project_id(self, project_id: str) -> "DataScienceModel"
        Sets the project ID.
    with_description(self, description: str) -> "DataScienceModel"
        Sets the description.
    with_compartment_id(self, compartment_id: str) -> "DataScienceModel"
        Sets the compartment ID.
    with_display_name(self, name: str) -> "DataScienceModel"
        Sets the name.
    with_freeform_tags(self, **kwargs: Dict[str, str]) -> "DataScienceModel"
        Sets freeform tags.
    with_defined_tags(self, **kwargs: Dict[str, Dict[str, object]]) -> "DataScienceModel"
        Sets defined tags.
    with_input_schema(self, schema: Union[Schema, Dict]) -> "DataScienceModel"
        Sets the model input schema.
    with_output_schema(self, schema: Union[Schema, Dict]) -> "DataScienceModel"
        Sets the model output schema.
    with_defined_metadata_list(self, metadata: Union[ModelTaxonomyMetadata, Dict]) -> "DataScienceModel"
        Sets model taxonomy (defined) metadata.
    with_custom_metadata_list(self, metadata: Union[ModelCustomMetadata, Dict]) -> "DataScienceModel"
        Sets model custom metadata.
    with_provenance_metadata(self, metadata: Union[ModelProvenanceMetadata, Dict]) -> "DataScienceModel"
        Sets model provenance metadata.
    with_artifact(self, uri: str)
        Sets the artifact location. Can be a local.
    with_model_version_set_id(self, model_version_set_id: str):
        Sets the model version set ID.
    with_version_label(self, version_label: str):
        Sets the model version label.


    Examples
    --------
    >>> ds_model = (DataScienceModel()
    ...    .with_compartment_id(os.environ["NB_SESSION_COMPARTMENT_OCID"])
    ...    .with_project_id(os.environ["PROJECT_OCID"])
    ...    .with_display_name("TestModel")
    ...    .with_description("Testing the test model")
    ...    .with_freeform_tags(tag1="val1", tag2="val2")
    ...    .with_artifact("/path/to/the/model/artifacts/"))
    >>> ds_model.create()
    >>> ds_model.status()
    >>> ds_model.with_description("new description").update()
    >>> ds_model.download_artifact("/path/to/dst/folder/")
    >>> ds_model.delete()
    >>> DataScienceModel.list()
    """

    _PREFIX = "datascience_model"

    CONST_ID = "id"
    CONST_PROJECT_ID = "projectId"
    CONST_COMPARTMENT_ID = "compartmentId"
    CONST_DISPLAY_NAME = "displayName"
    CONST_DESCRIPTION = "description"
    CONST_FREEFORM_TAG = "freeformTags"
    CONST_DEFINED_TAG = "definedTags"
    CONST_INPUT_SCHEMA = "inputSchema"
    CONST_OUTPUT_SCHEMA = "outputSchema"
    CONST_CUSTOM_METADATA = "customMetadataList"
    CONST_DEFINED_METADATA = "definedMetadataList"
    CONST_PROVENANCE_METADATA = "provenanceMetadata"
    CONST_ARTIFACT = "artifact"
    CONST_MODEL_VERSION_SET_ID = "modelVersionSetId"
    CONST_MODEL_VERSION_LABEL = "versionLabel"

    attribute_map = {
        CONST_ID: "id",
        CONST_PROJECT_ID: "project_id",
        CONST_COMPARTMENT_ID: "compartment_id",
        CONST_DISPLAY_NAME: "display_name",
        CONST_DESCRIPTION: "description",
        CONST_FREEFORM_TAG: "freeform_tags",
        CONST_DEFINED_TAG: "defined_tags",
        CONST_INPUT_SCHEMA: "input_schema",
        CONST_OUTPUT_SCHEMA: "output_schema",
        CONST_CUSTOM_METADATA: "custom_metadata_list",
        CONST_DEFINED_METADATA: "defined_metadata_list",
        CONST_PROVENANCE_METADATA: "provenance_metadata",
        CONST_ARTIFACT: "artifact",
        CONST_MODEL_VERSION_SET_ID: "model_version_set_id",
        CONST_MODEL_VERSION_LABEL: "version_label",
    }

    def __init__(self, spec: Dict = None, **kwargs) -> None:
        """Initializes datascience model.

        Parameters
        ----------
        spec: (Dict, optional). Defaults to None.
            Object specification.

        kwargs: Dict
            Specification as keyword arguments.
            If 'spec' contains the same key as the one in kwargs,
            the value from kwargs will be used.

            - project_id: str
            - compartment_id: str
            - name: str
            - description: str
            - defined_tags: Dict[str, Dict[str, object]]
            - freeform_tags: Dict[str, str]
            - input_schema: Union[ads.feature_engineering.Schema, Dict]
            - output_schema: Union[ads.feature_engineering.Schema, Dict]
            - defined_metadata_list: Union[ModelTaxonomyMetadata, Dict]
            - custom_metadata_list: Union[ModelCustomMetadata, Dict]
            - provenance_metadata: Union[ModelProvenanceMetadata, Dict]
            - artifact: str
        """
        super().__init__(spec=spec, **deepcopy(kwargs))
        # Reinitiate complex attributes
        self._init_complex_attributes()
        # Specify oci datascience model instance
        self.dsc_model = self._to_oci_dsc_model()

    @property
    def id(self) -> Optional[str]:
        """The model OCID."""
        if self.dsc_model:
            return self.dsc_model.id
        return None

    @property
    def status(self) -> Union[str, None]:
        """Status of the model.

        Returns
        -------
        str
            Status of the model.
        """
        if self.dsc_model:
            return self.dsc_model.status
        return None

    @property
    def kind(self) -> str:
        """The kind of the object as showing in a YAML."""
        return "datascienceModel"

    @property
    def project_id(self) -> str:
        return self.get_spec(self.CONST_PROJECT_ID)

    def with_project_id(self, project_id: str) -> "DataScienceModel":
        """Sets the project ID.

        Parameters
        ----------
        project_id: str
            The project ID.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        return self.set_spec(self.CONST_PROJECT_ID, project_id)

    @property
    def description(self) -> str:
        return self.get_spec(self.CONST_DESCRIPTION)

    def with_description(self, description: str) -> "DataScienceModel":
        """Sets the description.

        Parameters
        ----------
        description: str
            The description of the model.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        return self.set_spec(self.CONST_DESCRIPTION, description)

    @property
    def compartment_id(self) -> str:
        return self.get_spec(self.CONST_COMPARTMENT_ID)

    def with_compartment_id(self, compartment_id: str) -> "DataScienceModel":
        """Sets the compartment ID.

        Parameters
        ----------
        compartment_id: str
            The compartment ID.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        return self.set_spec(self.CONST_COMPARTMENT_ID, compartment_id)

    @property
    def display_name(self) -> str:
        return self.get_spec(self.CONST_DISPLAY_NAME)

    @display_name.setter
    def display_name(self, name: str) -> "DataScienceModel":
        return self.set_spec(self.CONST_DISPLAY_NAME, name)

    def with_display_name(self, name: str) -> "DataScienceModel":
        """Sets the name.

        Parameters
        ----------
        name: str
            The name.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        return self.set_spec(self.CONST_DISPLAY_NAME, name)

    @property
    def freeform_tags(self) -> Dict[str, str]:
        return self.get_spec(self.CONST_FREEFORM_TAG)

    def with_freeform_tags(self, **kwargs: Dict[str, str]) -> "DataScienceModel":
        """Sets freeform tags.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        return self.set_spec(self.CONST_FREEFORM_TAG, kwargs)

    @property
    def defined_tags(self) -> Dict[str, Dict[str, object]]:
        return self.get_spec(self.CONST_DEFINED_TAG)

    def with_defined_tags(
        self, **kwargs: Dict[str, Dict[str, object]]
    ) -> "DataScienceModel":
        """Sets defined tags.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        return self.set_spec(self.CONST_DEFINED_TAG, kwargs)

    @property
    def input_schema(self) -> Schema:
        """Returns model input schema.

        Returns
        -------
        ads.feature_engineering.Schema
            Model input schema.
        """
        return self.get_spec(self.CONST_INPUT_SCHEMA)

    def with_input_schema(self, schema: Union[Schema, Dict]) -> "DataScienceModel":
        """Sets the model input schema.

        Parameters
        ----------
        schema: Union[ads.feature_engineering.Schema, Dict]
            The model input schema.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        if schema and isinstance(schema, Dict):
            schema = Schema.from_dict(schema)
        return self.set_spec(self.CONST_INPUT_SCHEMA, schema)

    @property
    def output_schema(self) -> Schema:
        """Returns model output schema.

        Returns
        -------
        ads.feature_engineering.Schema
            Model output schema.
        """
        return self.get_spec(self.CONST_OUTPUT_SCHEMA)

    def with_output_schema(self, schema: Union[Schema, Dict]) -> "DataScienceModel":
        """Sets the model output schema.

        Parameters
        ----------
        schema: Union[ads.feature_engineering.Schema, Dict]
            The model output schema.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        if schema and isinstance(schema, Dict):
            schema = Schema.from_dict(schema)
        return self.set_spec(self.CONST_OUTPUT_SCHEMA, schema)

    @property
    def defined_metadata_list(self) -> ModelTaxonomyMetadata:
        """Returns model taxonomy (defined) metadatda."""
        return self.get_spec(self.CONST_DEFINED_METADATA)

    def with_defined_metadata_list(
        self, metadata: Union[ModelTaxonomyMetadata, Dict]
    ) -> "DataScienceModel":
        """Sets model taxonomy (defined) metadata.

        Parameters
        ----------
        metadata: Union[ModelTaxonomyMetadata, Dict]
            The defined metadata.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        if metadata and isinstance(metadata, Dict):
            metadata = ModelTaxonomyMetadata.from_dict(metadata)
        return self.set_spec(self.CONST_DEFINED_METADATA, metadata)

    @property
    def custom_metadata_list(self) -> ModelCustomMetadata:
        """Returns model custom metadatda."""
        return self.get_spec(self.CONST_CUSTOM_METADATA)

    def with_custom_metadata_list(
        self, metadata: Union[ModelCustomMetadata, Dict]
    ) -> "DataScienceModel":
        """Sets model custom metadata.

        Parameters
        ----------
        metadata: Union[ModelCustomMetadata, Dict]
            The custom metadata.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        if metadata and isinstance(metadata, Dict):
            metadata = ModelCustomMetadata.from_dict(metadata)
        return self.set_spec(self.CONST_CUSTOM_METADATA, metadata)

    @property
    def provenance_metadata(self) -> ModelProvenanceMetadata:
        """Returns model provenance metadatda."""
        return self.get_spec(self.CONST_PROVENANCE_METADATA)

    def with_provenance_metadata(
        self, metadata: Union[ModelProvenanceMetadata, Dict]
    ) -> "DataScienceModel":
        """Sets model provenance metadata.

        Parameters
        ----------
        provenance_metadata: Union[ModelProvenanceMetadata, Dict]
            The provenance metadata.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)
        """
        if metadata and isinstance(metadata, Dict):
            metadata = ModelProvenanceMetadata.from_dict(metadata)
        return self.set_spec(self.CONST_PROVENANCE_METADATA, metadata)

    @property
    def artifact(self) -> str:
        return self.get_spec(self.CONST_ARTIFACT)

    def with_artifact(self, uri: str):
        """Sets the artifact location. Can be a local.

        Parameters
        ----------
        uri: str
            Path to artifact directory or to the ZIP archive.
            It could contain a serialized model(required) as well as any files needed for deployment.
            The content of the source folder will be zipped and uploaded to the model catalog.

        Examples
        --------
        >>> .with_artifact(uri="./model1/")
        >>> .with_artifact(uri="./model1.zip")
        """
        return self.set_spec(self.CONST_ARTIFACT, uri)

    @property
    def model_version_set_id(self) -> str:
        return self.get_spec(self.CONST_MODEL_VERSION_SET_ID)

    def with_model_version_set_id(self, model_version_set_id: str):
        """Sets the model version set ID.

        Parameters
        ----------
        urmodel_version_set_idi: str
            The Model version set OCID.
        """
        return self.set_spec(self.CONST_MODEL_VERSION_SET_ID, model_version_set_id)

    @property
    def version_label(self) -> str:
        return self.get_spec(self.CONST_MODEL_VERSION_LABEL)

    def with_version_label(self, version_label: str):
        """Sets the model version label.

        Parameters
        ----------
        version_label: str
            The model version label.
        """
        return self.set_spec(self.CONST_MODEL_VERSION_LABEL, version_label)

    def create(self, **kwargs) -> "DataScienceModel":
        """Creates datascience model.

        Parameters
        ----------
        kwargs
            Additional kwargs arguments.
            Can be any attribute that `oci.data_science.models.Model` accepts.

            In addition can be also provided the attributes listed below.

            bucket_uri: (str, optional). Defaults to None.
                The OCI Object Storage URI where model artifacts will be copied to.
                The `bucket_uri` is only necessary for uploading large artifacts which
                size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
            overwrite_existing_artifact: (bool, optional). Defaults to `True`.
                Overwrite target bucket artifact if exists.
            remove_existing_artifact: (bool, optional). Defaults to `True`.
                Wether artifacts uploaded to object storage bucket need to be removed or not.
            region: (str, optional). Defaults to `None`.
                The destination Object Storage bucket region.
                By default the value will be extracted from the `OCI_REGION_METADATA` environment variable.
            auth: (Dict, optional). Defaults to `None`.
                The default authentication is set using `ads.set_auth` API.
                If you need to override the default, use the `ads.common.auth.api_keys` or
                `ads.common.auth.resource_principal` to create appropriate authentication signer
                and kwargs required to instantiate IdentityClient object.
            timeout: (int, optional). Defaults to 10 seconds.
                The connection timeout in seconds for the client.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self)

        Raises
        ------
        ValueError
            If compartment id not provided.
            If project id not provided.
        """
        if not self.compartment_id:
            raise ValueError("Compartment id must be provided.")

        if not self.project_id:
            raise ValueError("Project id must be provided.")

        if not self.display_name:
            self.display_name = self._random_display_name()

        payload = deepcopy(self._spec)
        payload.pop("id", None)
        logger.debug(f"Creating a model with payload {payload}")

        # Create model in the model catalog
        logger.info("Saving model to the Model Catalog.")
        self.dsc_model = self._to_oci_dsc_model(**kwargs).create()

        # Create model provenance
        if self.provenance_metadata:
            logger.info("Saving model provenance metadata.")
            self.dsc_model.create_model_provenance(
                self.provenance_metadata._to_oci_metadata()
            )

        # Upload artifacts
        logger.info("Uploading model artifacts.")
        self.upload_artifact(
            bucket_uri=kwargs.pop("bucket_uri", None),
            overwrite_existing_artifact=kwargs.pop("overwrite_existing_artifact", True),
            remove_existing_artifact=kwargs.pop("remove_existing_artifact", True),
            region=kwargs.pop("region", None),
            auth=kwargs.pop("auth", None),
            timeout=kwargs.pop("timeout", None),
        )

        # Sync up model
        self.sync()
        logger.info(f"Model {self.id} has been successfully saved.")

        return self

    def upload_artifact(
        self,
        bucket_uri: Optional[str] = None,
        auth: Optional[Dict] = None,
        region: Optional[str] = None,
        overwrite_existing_artifact: Optional[bool] = True,
        remove_existing_artifact: Optional[bool] = True,
        timeout: Optional[int] = None,
    ) -> None:
        """Uploads model artifacts to the model catalog.

        Parameters
        ----------
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for uploading large artifacts which
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        auth: (Dict, optional). Defaults to `None`.
            The default authentication is set using `ads.set_auth` API.
            If you need to override the default, use the `ads.common.auth.api_keys` or
            `ads.common.auth.resource_principal` to create appropriate authentication signer
            and kwargs required to instantiate IdentityClient object.
        region: (str, optional). Defaults to `None`.
            The destination Object Storage bucket region.
            By default the value will be extracted from the `OCI_REGION_METADATA` environment variables.
        overwrite_existing_artifact: (bool, optional). Defaults to `True`.
            Overwrite target bucket artifact if exists.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Wether artifacts uploaded to object storage bucket need to be removed or not.
        timeout: (int, optional). Defaults to 10 seconds.
            The connection timeout in seconds for the client.
        """
        # Upload artifact to the model catalog
        if not self.artifact:
            logger.warn(
                "Model artifact location not provided. "
                "Provide the artifact location to upload artifacts to the model catalog."
            )
            return

        if timeout:
            self.dsc_model._client = None
            self.dsc_model.__class__.kwargs = {
                **(self.dsc_model.__class__.kwargs or {}),
                "timeout": timeout,
            }

        if bucket_uri or utils.folder_size(self.artifact) > _MAX_ARTIFACT_SIZE_IN_BYTES:
            if not bucket_uri:
                raise ModelArtifactSizeError(
                    max_artifact_size=utils.human_size(_MAX_ARTIFACT_SIZE_IN_BYTES)
                )

            artifact_uploader = LargeArtifactUploader(
                dsc_model=self.dsc_model,
                artifact_path=self.artifact,
                auth=auth,
                region=region,
                bucket_uri=bucket_uri,
                overwrite_existing_artifact=overwrite_existing_artifact,
                remove_existing_artifact=remove_existing_artifact,
            )
        else:
            artifact_uploader = SmallArtifactUploader(
                dsc_model=self.dsc_model,
                artifact_path=self.artifact,
            )

        artifact_uploader.upload()

    def download_artifact(
        self,
        target_dir: str,
        auth: Optional[Dict] = None,
        force_overwrite: Optional[bool] = False,
        bucket_uri: Optional[str] = None,
        region: Optional[str] = None,
        overwrite_existing_artifact: Optional[bool] = True,
        remove_existing_artifact: Optional[bool] = True,
        timeout: Optional[int] = None,
    ):
        """Downloads model artifacts from the model catalog.

        Parameters
        ----------
        target_dir: str
            The target location of model artifacts.
        auth: (Dict, optional). Defaults to `None`.
            The default authentication is set using `ads.set_auth` API.
            If you need to override the default, use the `ads.common.auth.api_keys` or
            `ads.common.auth.resource_principal` to create appropriate authentication signer
            and kwargs required to instantiate IdentityClient object.
        force_overwrite: (bool, optional). Defaults to `False`.
            Overwrite target directory if exists.
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for uploading large artifacts which
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        region: (str, optional). Defaults to `None`.
            The destination Object Storage bucket region.
            By default the value will be extracted from the `OCI_REGION_METADATA` environment variables.
        overwrite_existing_artifact: (bool, optional). Defaults to `True`.
            Overwrite target bucket artifact if exists.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Wether artifacts uploaded to object storage bucket need to be removed or not.
        timeout: (int, optional). Defaults to 10 seconds.
            The connection timeout in seconds for the client.

        Raises
        ------
        ModelArtifactSizeError
            If model artifacts size greater than 2GB and temporary OS bucket uri not provided.
        """
        # Upload artifact to the model catalog
        if not self.artifact:
            logger.warn(
                "Model doesn't contain an artifact. "
                "The artifact needs to be uploaded to the model catalog at first. "
            )
            return

        if timeout:
            self.dsc_model._client = None
            self.dsc_model.__class__.kwargs = {
                **(self.dsc_model.__class__.kwargs or {}),
                "timeout": timeout,
            }

        artifact_info = self.dsc_model.get_artifact_info()
        artifact_size = int(artifact_info.get("content-length"))
        if not bucket_uri and artifact_size > _MAX_ARTIFACT_SIZE_IN_BYTES:
            raise ModelArtifactSizeError(utils.human_size(_MAX_ARTIFACT_SIZE_IN_BYTES))

        if artifact_size > _MAX_ARTIFACT_SIZE_IN_BYTES or bucket_uri:
            artifact_downloader = LargeArtifactDownloader(
                dsc_model=self.dsc_model,
                target_dir=target_dir,
                auth=auth,
                force_overwrite=force_overwrite,
                region=region,
                bucket_uri=bucket_uri,
                overwrite_existing_artifact=overwrite_existing_artifact,
                remove_existing_artifact=remove_existing_artifact,
            )
        else:
            artifact_downloader = SmallArtifactDownloader(
                dsc_model=self.dsc_model,
                target_dir=target_dir,
                force_overwrite=force_overwrite,
            )

        artifact_downloader.download()

    def update(self, **kwargs) -> "DataScienceModel":
        """Updates datascience model in model catalog.

        Parameters
        ----------
        kwargs
            Additional kwargs arguments.
            Can be any attribute that `oci.data_science.models.Model` accepts.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self).
        """
        if not self.id:
            logger.warn(
                "Model needs to be saved to the model catalog before it can be updated."
            )
            return

        logger.debug(f"Updating a model with payload {self._spec}")
        logger.info(f"Updating model {self.id} in the Model Catalog.")
        self.dsc_model = self._to_oci_dsc_model(**kwargs).update()

        logger.debug(f"Updating a model provenance metadata {self.provenance_metadata}")
        try:
            self.dsc_model.get_model_provenance()
            self.dsc_model.update_model_provenance(
                self.provenance_metadata._to_oci_metadata()
            )
        except ModelProvenanceNotFoundError:
            self.dsc_model.create_model_provenance(
                self.provenance_metadata._to_oci_metadata()
            )

        return self.sync()

    def delete(
        self,
        delete_associated_model_deployment: Optional[bool] = False,
    ) -> "DataScienceModel":
        """Removes model from the model catalog.

        Parameters
        ----------
        delete_associated_model_deployment: (bool, optional). Defaults to `False`.
            Whether associated model deployments need to be deleted or not.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self).
        """
        self.dsc_model.delete(delete_associated_model_deployment)
        return self.sync()

    @classmethod
    def list(
        cls, compartment_id: str = None, project_id: str = None, **kwargs
    ) -> List["DataScienceModel"]:
        """Lists datascience models in a given compartment.

        Parameters
        ----------
        compartment_id: (str, optional). Defaults to `None`.
            The compartment OCID.
        project_id: (str, optional). Defaults to `None`.
            The project OCID.
        kwargs
            Additional keyword arguments for filtering models.

        Returns
        -------
        List[DataScienceModel]
            The list of the datascience models.
        """
        return [
            cls()._update_from_oci_dsc_model(model)
            for model in OCIDataScienceModel.list_resource(
                compartment_id, project_id=project_id, **kwargs
            )
        ]

    @classmethod
    def list_df(
        cls, compartment_id: str = None, project_id: str = None, **kwargs
    ) -> "pandas.DataFrame":
        """Lists datascience models in a given compartment.

        Parameters
        ----------
        compartment_id: (str, optional). Defaults to `None`.
            The compartment OCID.
        project_id: (str, optional). Defaults to `None`.
            The project OCID.
        kwargs
            Additional keyword arguments for filtering models.

        Returns
        -------
        pandas.DataFrame
            The list of the datascience models in a pandas dataframe format.
        """
        records = []
        for model in OCIDataScienceModel.list_resource(
            compartment_id, project_id=project_id, **kwargs
        ):
            records.append(
                {
                    "id": f"...{model.id[-6:]}",
                    "display_name": model.display_name,
                    "description": model.description,
                    "time_created": model.time_created.strftime(utils.date_format),
                    "lifecycle_state": model.lifecycle_state,
                    "created_by": f"...{model.created_by[-6:]}",
                    "compartment_id": f"...{model.compartment_id[-6:]}",
                    "project_id": f"...{model.project_id[-6:]}",
                }
            )
        return pandas.DataFrame.from_records(records)

    @classmethod
    def from_id(cls, id: str) -> "DataScienceModel":
        """Gets an existing model by OCID.

        Parameters
        ----------
        id: str
            The model OCID.

        Returns
        -------
        DataScienceModel
            An instance of DataScienceModel.
        """
        return cls()._update_from_oci_dsc_model(OCIDataScienceModel.from_id(id))

    def sync(self):
        """Sync up a datascience model with OCI datascience model."""
        return self._update_from_oci_dsc_model(OCIDataScienceModel.from_id(self.id))

    def _init_complex_attributes(self):
        """Initiates complex attributes."""
        self.with_custom_metadata_list(self.custom_metadata_list)
        self.with_defined_metadata_list(self.defined_metadata_list)
        self.with_provenance_metadata(self.provenance_metadata)
        self.with_input_schema(self.input_schema)
        self.with_output_schema(self.output_schema)

    def _to_oci_dsc_model(self, **kwargs):
        """Creates an `OCIDataScienceModel` instance from the  `DataScienceModel`.

        kwargs
            Additional kwargs arguments.
            Can be any attribute that `oci.data_science.models.Model` accepts.

        Returns
        -------
        OCIDataScienceModel
            The instance of the OCIDataScienceModel.
        """
        COMPLEX_ATTRIBUTES_CONVERTER = {
            self.CONST_INPUT_SCHEMA: "to_json",
            self.CONST_OUTPUT_SCHEMA: "to_json",
            self.CONST_CUSTOM_METADATA: "_to_oci_metadata",
            self.CONST_DEFINED_METADATA: "_to_oci_metadata",
            self.CONST_PROVENANCE_METADATA: "_to_oci_metadata",
        }
        dsc_spec = {}
        for infra_attr, dsc_attr in self.attribute_map.items():
            value = self.get_spec(infra_attr)
            if infra_attr in COMPLEX_ATTRIBUTES_CONVERTER and value:
                dsc_spec[dsc_attr] = getattr(
                    self.get_spec(infra_attr), COMPLEX_ATTRIBUTES_CONVERTER[infra_attr]
                )()
            else:
                dsc_spec[dsc_attr] = value

        dsc_spec.update(**kwargs)
        return OCIDataScienceModel(**dsc_spec)

    def _update_from_oci_dsc_model(
        self, dsc_model: OCIDataScienceModel
    ) -> "DataScienceModel":
        """Update the properties from an OCIDataScienceModel object.

        Parameters
        ----------
        dsc_model: OCIDataScienceModel
            An instance of OCIDataScienceModel.

        Returns
        -------
        DataScienceModel
            The DataScienceModel instance (self).
        """
        COMPLEX_ATTRIBUTES_CONVERTER = {
            self.CONST_INPUT_SCHEMA: Schema.from_json,
            self.CONST_OUTPUT_SCHEMA: Schema.from_json,
            self.CONST_CUSTOM_METADATA: ModelCustomMetadata._from_oci_metadata,
            self.CONST_DEFINED_METADATA: ModelTaxonomyMetadata._from_oci_metadata,
        }

        # Update the main properties
        self.dsc_model = dsc_model
        for infra_attr, dsc_attr in self.attribute_map.items():
            value = utils.get_value(dsc_model, dsc_attr)
            if value:
                if infra_attr in COMPLEX_ATTRIBUTES_CONVERTER:
                    value = COMPLEX_ATTRIBUTES_CONVERTER[infra_attr](value)
                self.set_spec(infra_attr, value)

        # Update provenance metadata
        try:
            self.set_spec(
                self.CONST_PROVENANCE_METADATA,
                ModelProvenanceMetadata._from_oci_metadata(
                    self.dsc_model.get_model_provenance()
                ),
            )
        except ModelProvenanceNotFoundError:
            pass

        # Update artifact info
        try:
            artifact_info = self.dsc_model.get_artifact_info()
            _, file_name_info = cgi.parse_header(artifact_info["Content-Disposition"])
            self.set_spec(self.CONST_ARTIFACT, file_name_info["filename"])
        except:
            pass

        return self

    def to_dict(self) -> Dict:
        """Serializes model to a dictionary.

        Returns
        -------
        dict
            The model serialized as a dictionary.
        """
        spec = deepcopy(self._spec)
        for key, value in spec.items():
            if hasattr(value, "to_dict"):
                value = value.to_dict()
            spec[key] = value

        return {
            "kind": self.kind,
            "type": self.type,
            "spec": utils.batch_convert_case(spec, "camel"),
        }

    @classmethod
    def from_dict(cls, config: Dict) -> "DataScienceModel":
        """Loads model instance from a dictionary of configurations.

        Parameters
        ----------
        config: Dict
            A dictionary of configurations.

        Returns
        -------
        DataScienceModel
            The model instance.
        """
        return cls(spec=utils.batch_convert_case(deepcopy(config["spec"]), "snake"))

    def _random_display_name(self):
        """Generates a random display name."""
        return f"{self._PREFIX}-{utils.get_random_name_for_resource()}"

    def _load_default_properties(self) -> Dict:
        """Load default properties from environment variables, notebook session, etc.

        Returns
        -------
        Dict
            A dictionary of default properties.
        """
        defaults = super()._load_default_properties()
        compartment_ocid = COMPARTMENT_OCID
        if compartment_ocid:
            defaults[self.CONST_COMPARTMENT_ID] = compartment_ocid
        if PROJECT_OCID:
            defaults[self.CONST_PROJECT_ID] = PROJECT_OCID
        defaults[self.CONST_DISPLAY_NAME] = self._random_display_name()

        return defaults

    def __getattr__(self, item):
        if f"with_{item}" in self.__dir__():
            return self.get_spec(item)
        raise AttributeError(f"Attribute {item} not found.")



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import fnmatch
import importlib
import os
import sys
import shutil
import tempfile
import uuid
from typing import Dict, Optional, Tuple
from ads.common import auth as authutil
from ads.common import logger, utils
from ads.config import CONDA_BUCKET_NAME, CONDA_BUCKET_NS
from ads.model.runtime.env_info import EnvInfo, InferenceEnvInfo, TrainingEnvInfo
from ads.model.runtime.runtime_info import RuntimeInfo
from jinja2 import Environment, PackageLoader
import warnings
from ads import __version__
from datetime import datetime

MODEL_ARTIFACT_VERSION = "3.0"
REQUIRED_ARTIFACT_FILES = ("runtime.yaml", "score.py")
SCORE_VERSION = "1.0"
ADS_VERSION = __version__


class ArtifactNestedFolderError(Exception):
    def __init__(self, folder: str):
        self.folder = folder
        super().__init__("The required artifact files placed in a nested folder.")


class ArtifactRequiredFilesError(Exception):
    def __init__(self, required_files: Tuple[str]):
        super().__init__(
            "Not all required files presented in artifact folder. "
            f"Required files: {required_files}"
        )


class AritfactFolderStructureError(Exception):
    def __init__(self, required_files: Tuple[str]):
        super().__init__(
            "The artifact folder has a wrong structure. "
            f"Required files: {required_files}"
        )


def _validate_artifact_dir(
    artifact_dir: str, required_files: Tuple[str] = REQUIRED_ARTIFACT_FILES
) -> bool:
    """The function helper to validate artifacts folder structure.

    Params
    ------
        artifact_dir: str
            The local artifact folder to store the files needed for deployment.
        required_files: (Tuple[str], optional). Defaults to ("runtime.yaml", "score.py").
            The list of required artifact files.

    Raises:
        ValueError
            If `required_files` not provided.
            If `artifact_dir` not exists.
        ArtifactNestedFolderError
            If artifact files located in a nested folder.
        ArtifactRequiredFilesError
            If not all required files found in artifact folder.
        AritfactFolderStructureError
            In case if artifact folder has a wrong structure.

    Returns:
        bool: True if artifact folder contains the list of the all required files.
    """
    if not required_files or len(required_files) == 0:
        raise ValueError("Required artifact files not provided.")

    artifact_dir = os.path.abspath(os.path.expanduser(artifact_dir))
    if not os.path.exists(artifact_dir):
        raise ValueError(f"The path `{artifact_dir}` not found.")

    result = {required_file.lower(): None for required_file in required_files}
    for dirpath, _, filenames in os.walk(artifact_dir):
        rel_path = os.path.abspath(dirpath)
        for required_file in required_files:
            for filename in fnmatch.filter(filenames, required_file):
                if filename.lower() in result and result[filename] == None:
                    result[filename] = rel_path

    # if not required artifact files found in provided artifact dir
    if None in result.values():
        raise ArtifactRequiredFilesError(required_files)
    # if required artifact files placed in different nested folders
    if len(set(result.values())) > 1:
        raise AritfactFolderStructureError(required_files)

    if all(path == artifact_dir for path in result.values()):
        return True

    # if required files are placed in a nested folder
    raise (ArtifactNestedFolderError(list(result.values())[0]))


class ModelArtifact:
    """The class that represents model artifacts.
    It is designed to help to generate and manage model artifacts.
    """

    def __init__(
        self,
        artifact_dir: str,
        model_file_name: str = None,
        reload: Optional[bool] = False,
    ):
        """Initializes a ModelArtifact instance.

        Parameters
        ----------
        artifact_dir: str
            The local artifact folder to store the files needed for deployment.
        model_file_name: (str, optional). Defaults to `None`.
            The file name of the serialized model.
        reload: (bool, optional). Defaults to False.
            Determine whether will reload the Model into the env.

        Returns
        -------
        ModelArtifact
            A ModelArtifact instance.

        Raises
        ------
        ValueError
            If `artifact_dir` not provided.
        """
        if not artifact_dir:
            raise ValueError("The `artifact_dir` needs to be provided.")

        self.score = None
        self.artifact_dir = os.path.abspath(os.path.expanduser(artifact_dir))
        sys.path.insert(0, self.artifact_dir)
        self.model_file_name = model_file_name
        self._env = Environment(loader=PackageLoader("ads", "templates"))
        if reload:
            self.reload()
            # Extracts the model_file_name from the score.py.
            if (
                not self.model_file_name
                and self.score
                and hasattr(self.score, "model_name")
                and self.score.model_name
            ):
                self.model_file_name = self.score.model_name

    def prepare_runtime_yaml(
        self,
        inference_conda_env: str,
        inference_python_version: str = None,
        training_conda_env: str = None,
        training_python_version: str = None,
        force_overwrite: bool = False,
        namespace: str = CONDA_BUCKET_NS,
        bucketname: str = CONDA_BUCKET_NAME,
        auth: dict = None,
    ) -> None:
        """Generate a runtime yaml file and save it to the artifact
        directory.

        Parameters
        ----------
        inference_conda_env: (str, optional). Defaults to None.
            The object storage path of conda pack which will be used in deployment.
            Can be either slug or object storage path of the conda pack.
            You can only pass in slugs if the conda pack is a service pack.
        inference_python_version: (str, optional). Defaults to None.
            The python version which will be used in deployment.
        training_conda_env: (str, optional). Defaults to None.
            The object storage path of conda pack used during training.
            Can be either slug or object storage path of the conda pack.
            You can only pass in slugs if the conda pack is a service pack.
        training_python_version: (str, optional). Defaults to None.
            The python version used during training.
        force_overwrite : (bool, optional). Defaults to False.
            Whether to overwrite existing files.
        namespace: (str, optional)
            The namespace of region. Defaults to environment variable CONDA_BUCKET_NS.
        bucketname: (str, optional)
            The bucketname of service pack. Defaults to environment variable CONDA_BUCKET_NAME.
        auth :(Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Raises
        ------
        ValueError
            If neither slug or conda_env_uri is provided.

        Returns
        -------
        RuntimeInfo
            A RuntimeInfo instance.
        """
        runtime_info = RuntimeInfo.from_env()
        runtime_info.model_artifact_version = MODEL_ARTIFACT_VERSION
        inference_conda_env = ModelArtifact._populate_env_info(
            InferenceEnvInfo,
            conda_pack=inference_conda_env,
            bucketname=bucketname,
            namespace=namespace,
            auth=auth,
        )

        if training_conda_env:
            training_conda_env = ModelArtifact._populate_env_info(
                TrainingEnvInfo,
                conda_pack=training_conda_env,
                bucketname=bucketname,
                namespace=namespace,
                auth=auth,
            )
        else:
            training_conda_env = TrainingEnvInfo()
        if training_python_version:
            training_conda_env.training_python_version = training_python_version
        if inference_python_version:
            inference_conda_env.inference_python_version = inference_python_version
        runtime_info.model_deployment.inference_conda_env = inference_conda_env
        runtime_info.model_provenance.training_conda_env = training_conda_env
        runtime_info.model_provenance.training_code.artifact_directory = (
            self.artifact_dir
        )
        if (
            not runtime_info.model_deployment.inference_conda_env.inference_python_version
            or runtime_info.model_deployment.inference_conda_env.inference_python_version.strip()
            == ""
        ):
            warnings.warn(
                "Cannot automatically detect the inference python version. `inference_python_version` must be provided."
            )
        runtime_file_path = os.path.join(self.artifact_dir, "runtime.yaml")
        if os.path.exists(runtime_file_path) and not force_overwrite:
            raise ValueError(
                "runtime.yaml already exists. "
                "Set `force_overwrite` to True to overwrite all the files."
            )
        else:
            runtime_info.save()
        return runtime_info

    @staticmethod
    def _populate_env_info(
        clss: EnvInfo,
        conda_pack: str,
        bucketname: str = None,
        namespace: str = None,
        auth: dict = None,
    ) -> "EnvInfo":
        """Populates the Training/InferenceEnvInfo instance.

        Parameters
        ----------
        clss: EnvInfo
            A EnvInfo class.
        conda_pack: str
            The object storage path of conda pack.
            Can be either slug or object storage path of the conda pack.
            You can only pass in slugs if the conda pack is a service pack.
        namespace: (str, optional)
            The namespace of region.
        bucketname: (str, optional)
            The bucketname of service pack.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.

        Returns
        -------
        EnvInfo
            An EnvInfo instance.
        """
        if conda_pack.startswith("oci://"):
            return clss.from_path(conda_pack)
        return clss.from_slug(
            env_slug=conda_pack, bucketname=bucketname, namespace=namespace, auth=auth
        )

    def prepare_score_py(
        self, jinja_template_filename: str, model_file_name: str = None, **kwargs
    ):
        """Prepares `score.py` file.

        Parameters
        ----------
        jinja_template_filename: str.
            The jinja template file name.
        model_file_name: (str, optional). Defaults to `None`.
            The file name of the serialized model.
        **kwargs: (dict)
            use_torch_script: bool
            data_deserializer: str

        Returns
        -------
        None

        Raises
        ------
        ValueError
            If `model_file_name` not provided.
        """
        self.model_file_name = model_file_name or self.model_file_name
        if not self.model_file_name:
            raise ValueError("The `model_file_name` must be provided.")

        if not os.path.exists(
            os.path.join(
                os.path.abspath(os.path.join(os.path.dirname(__file__), "..")),
                "templates",
                f"{jinja_template_filename}.jinja2",
            )
        ):
            raise FileExistsError(f"{jinja_template_filename}.jinja2 does not exists.")
        scorefn_template = self._env.get_template(f"{jinja_template_filename}.jinja2")
        time_suffix = datetime.today().strftime("%Y%m%d_%H%M%S")

        context = {
            "model_file_name": self.model_file_name,
            "SCORE_VERSION": SCORE_VERSION,
            "ADS_VERSION": ADS_VERSION,
            "time_created": time_suffix,
        }
        context.update(kwargs)
        with open(os.path.join(self.artifact_dir, "score.py"), "w") as sfl:
            sfl.write(scorefn_template.render(context))

    def reload(self):
        """Syncs the `score.py` to reload the model and predict function.

        Returns
        -------
        None
            Nothing
        """
        spec = importlib.util.spec_from_file_location(
            "score%s" % uuid.uuid4(), os.path.join(self.artifact_dir, "score.py")
        )
        self.score = importlib.util.module_from_spec(spec)
        spec.loader.exec_module(self.score)
        self.model = self.score.load_model()  # load model in cache
        # remove the cache files.
        for dir in [
            os.path.join(self.artifact_dir, "__pycache__"),
            os.path.join(self.artifact_dir, ".ipynb_checkpoints"),
        ]:
            if os.path.exists(dir):
                shutil.rmtree(dir, ignore_errors=True)

    @classmethod
    def from_uri(
        cls,
        uri: str,
        artifact_dir: str,
        model_file_name: str = None,
        force_overwrite: Optional[bool] = False,
        auth: Optional[Dict] = None,
    ):
        """Constructs a ModelArtifact object from the existing model artifacts.

        Parameters
        ----------
        uri: str
            The URI of source artifact folder or achive. Can be local path or
            OCI object storage URI.
        artifact_dir: str
            The local artifact folder to store the files needed for deployment.
        model_file_name: (str, optional). Defaults to `None`
            The file name of the serialized model.
        force_overwrite: (bool, optional). Defaults to False.
            Whether to overwrite existing files or not.
        auth: (Dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API.
            If you need to override the default, use the `ads.common.auth.api_keys`
            or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate
            IdentityClient object.

        Returns
        -------
        ModelArtifact
            A `ModelArtifact` instance

        Raises
        ------
        ValueError
            If `uri` is equal to `artifact_dir`, and it not exists.
        """
        artifact_dir = os.path.join(
            os.path.abspath(os.path.expanduser(artifact_dir)), ""
        )

        if artifact_dir == os.path.join(
            os.path.abspath(os.path.expanduser(uri)).rstrip("/"), ""
        ):
            if not os.path.exists(artifact_dir):
                raise ValueError("Provided `uri` doesn't exist.")
        else:
            auth = auth or authutil.default_signer()
            utils.copy_from_uri(
                uri=uri,
                to_path=artifact_dir,
                unpack=True,
                force_overwrite=force_overwrite,
                auth=auth,
            )
        try:
            _validate_artifact_dir(artifact_dir)
        except ArtifactNestedFolderError as exc:
            with tempfile.TemporaryDirectory() as temp_dir:
                utils.copy_from_uri(
                    uri=exc.folder, to_path=temp_dir, force_overwrite=True
                )
                utils.copy_from_uri(
                    uri=temp_dir, to_path=artifact_dir, force_overwrite=True
                )

        return cls(artifact_dir, model_file_name, reload=True)

    def __getattr__(self, item):
        """Makes the functions in `score.py` directly accessable by ModelArtifact class."""
        return getattr(self.score, item)



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2022 Oracle and its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""
The module that provides the decorator helping to convert function arguments
to specific case (lower/upper).

Examples
--------
>>> @argument_to_case("lower", ["name"])
... def test_function(name: str, last_name: str)
...     print(name)
...
>>> test_function("myname", "mylastname")
... MYNAME
"""

import inspect
from copy import copy
from enum import Enum
from functools import wraps
from typing import Callable, List

from .utils import _get_original_func


class ArgumentCase(Enum):
    LOWER = "lower"
    UPPER = "upper"


def argument_to_case(case: ArgumentCase, arguments: List[str]):
    """The decorator helping to convert function arguments to specific case.

    Parameters
    ----------
    case: ArgumentCase
        The case to convert specific arguments.
    arguments: List[str]
        The list of arguments to convert to specific case.

    Examples
    --------
    >>> @argument_to_case("lower", ["name"])
    ... def test_function(name: str, last_name: str)
    ...     print(name)
    ...
    >>> test_function("myname", "mylastname")
    ... MYNAME
    """

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # Retrieving the original function from the decorated one.
            # This is necessary when the chain of decorators is used.
            # The only original function arguments should be processed.
            original_func = _get_original_func(func)
            # Getting the original function arguments
            func_args = inspect.getfullargspec(original_func).args
            # Saving original args and kwargs.
            new_args = list(copy(args))
            new_kwargs = copy(kwargs)
            # Converting args and kwargs to the specific case.
            for func_arg in func_args:
                if func_arg in arguments:
                    if func_arg in new_kwargs:
                        new_kwargs[func_arg] = (
                            new_kwargs[func_arg].lower()
                            if case == ArgumentCase.LOWER
                            else new_kwargs[func_arg].upper()
                        )
                    else:
                        arg_index = func_args.index(func_arg)
                        if arg_index >= 0:
                            new_args[arg_index] = (
                                new_args[arg_index].lower()
                                if case == ArgumentCase.LOWER
                                else new_args[arg_index].upper()
                            )
            return func(*new_args, **new_kwargs)

        return wrapper

    return decorator



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import warnings

warnings.warn(
    (
        "The `ads.catalog.model` is deprecated in `oracle-ads 2.6.9` and will be removed in `oracle-ads 3.0`. "
        "Use framework specific Model utility class for saving and deploying model. "
        "Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html"
    ),
    DeprecationWarning,
    stacklevel=2,
)

import json
import os
import shutil
import tempfile
import time
import uuid
from typing import Dict, Optional, Union
from zipfile import ZipFile

import pandas as pd
import yaml
from ads.catalog.summary import SummaryList
from ads.common import auth, logger, oci_client, utils
from ads.common.decorator.deprecate import deprecated
from ads.common.decorator.runtime_dependency import (
    runtime_dependency,
    OptionalDependency,
)
from ads.common.model_artifact import ConflictStrategy, ModelArtifact
from ads.model.model_metadata import (
    METADATA_SIZE_LIMIT,
    MetadataSizeTooLarge,
    ModelCustomMetadata,
    ModelTaxonomyMetadata,
)
from ads.common.object_storage_details import ObjectStorageDetails
from ads.common.oci_resource import SEARCH_TYPE, OCIResource
from ads.config import (
    NB_SESSION_COMPARTMENT_OCID,
    OCI_ODSC_SERVICE_ENDPOINT,
    OCI_REGION_METADATA,
    PROJECT_OCID,
)
from ads.dataset.progress import TqdmProgressBar
from ads.feature_engineering.schema import Schema
from ads.model.model_version_set import ModelVersionSet, _extract_model_version_set_id
from ads.model.deployment.model_deployer import ModelDeployer
from oci.data_science.data_science_client import DataScienceClient
from oci.data_science.models import (
    ArtifactExportDetailsObjectStorage,
    ArtifactImportDetailsObjectStorage,
    CreateModelDetails,
    ExportModelArtifactDetails,
    ImportModelArtifactDetails,
)
from oci.data_science.models import Model as OCIModel
from oci.data_science.models import ModelSummary, WorkRequest
from oci.data_science.models.model_provenance import ModelProvenance
from oci.data_science.models.update_model_details import UpdateModelDetails
from oci.exceptions import ServiceError
from oci.identity import IdentityClient

_UPDATE_MODEL_DETAILS_ATTRIBUTES = [
    "display_name",
    "description",
    "freeform_tags",
    "defined_tags",
    "model_version_set_id",
    "version_label",
]
_MODEL_PROVENANCE_ATTRIBUTES = ModelProvenance().swagger_types.keys()
_ETAG_KEY = "ETag"
_MAX_ARTIFACT_SIZE_IN_BYTES = 2147483648  # 2GB
_WORK_REQUEST_INTERVAL_IN_SEC = 3


class ModelWithActiveDeploymentError(Exception):
    pass


class ModelArtifactSizeError(Exception):
    def __init__(self, max_artifact_size: str):
        super().__init__(
            f"The model artifacts size is greater than `{max_artifact_size}`. "
            "The `bucket_uri` needs to be specified to "
            "copy artifacts to the object storage bucket. "
            "Example: `bucket_uri=oci://<bucket_name>@<namespace>/prefix/`"
        )


def _get_etag(response) -> str:
    """Gets etag from the response."""
    if _ETAG_KEY in response.headers:
        return response.headers[_ETAG_KEY].split("--")[0]
    return None


class ModelSummaryList(SummaryList):
    """Model Summary List which represents a list of Model Object.

    Methods
    -------
    sort_by(self, columns, reverse=False)
        Performs a multi-key sort on a particular set of columns and returns the sorted ModelSummaryList.
        Results are listed in a descending order by default.
    filter(self, selection, instance=None)
        Filters the model list according to a lambda filter function, or list comprehension.
    """

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def __init__(
        self,
        model_catalog,
        model_list,
        response=None,
        datetime_format=utils.date_format,
    ):
        super(ModelSummaryList, self).__init__(
            model_list, datetime_format=datetime_format
        )
        self.mc = model_catalog
        self.response = response

    def __add__(self, rhs):
        return ModelSummaryList(
            self.mc, list.__add__(self, rhs), datetime_format=self.datetime_format
        )

    def __getitem__(self, item):
        return self.mc.get_model(super(ModelSummaryList, self).__getitem__(item).id)

    def sort_by(self, columns, reverse=False):
        """
        Performs a multi-key sort on a particular set of columns and returns the sorted ModelSummaryList.
        Results are listed in a descending order by default.

        Parameters
        ----------
        columns: List of string
          A list of columns which are provided to sort on
        reverse: Boolean (defaults to false)
          If you'd like to reverse the results (for example, to get ascending instead of descending results)

        Returns
        -------
        ModelSummaryList: A sorted ModelSummaryList
        """
        return ModelSummaryList(
            self.mc,
            self._sort_by(columns, reverse=reverse),
            datetime_format=self.datetime_format,
        )

    def filter(self, selection, instance=None):
        """
        Filters the model list according to a lambda filter function, or list comprehension.

        Parameters
        ----------
        selection: lambda function filtering model instances, or a list-comprehension
            function of list filtering projects
        instance: list, optional
            list to filter, optional, defaults to self

        Returns
        -------
        ModelSummaryList: A filtered ModelSummaryList
        """
        instance = instance if instance is not None else self

        if callable(selection):
            res = list(filter(selection, instance))
            # lambda filtering
            if len(res) == 0:
                print("No models found")
                return
            return ModelSummaryList(self.mc, res, datetime_format=self.datetime_format)
        elif isinstance(selection, list):
            # list comprehension
            if len(selection) == 0:
                print("No models found")
                return
            return ModelSummaryList(
                self.mc, selection, datetime_format=self.datetime_format
            )
        else:
            raise ValueError(
                "Filter selection must be a function or a ProjectSummaryList"
            )


class Model:
    """Class that represents the ADS implementation of model catalog item.
    Converts the metadata and schema from OCI implememtation to ADS implementation.

    Methods
    -------
    to_dataframe
        Converts model to dataframe format.
    show_in_notebook
        Shows model in the notebook in dataframe or YAML representation.
    activate
        Activates model.
    deactivate
        Deactivates model.
    commit
        Commits the changes made to the model.
    rollback
        Rollbacks the changes made to the model.
    load_model
        Loads the model from the model catalog based on model ID.
    """

    _FIELDS_TO_DECORATE = [
        "schema_input",
        "schema_output",
        "metadata_custom",
        "metadata_taxonomy",
    ]

    _NEW_ATTRIBUTES = [
        "input_schema",
        "output_schema",
        "custom_metadata_list",
        "defined_metadata_list",
    ]

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def __init__(
        self,
        model: OCIModel,
        model_etag: str,
        provenance_metadata: ModelProvenance,
        provenance_etag: str,
        ds_client: DataScienceClient,
        identity_client: IdentityClient,
    ) -> None:
        """Initializes the Model.

        Parameters
        ----------
        model: OCIModel
            The OCI model object.
        model_etag: str
            The model ETag.
        provenance_metadata: ModelProvenance
            The model provenance metadata.
        provenance_etag: str
            The model provenance metadata ETag.
        ds_client: DataScienceClient
            The Oracle DataScience client.
        identity_client: IdentityClient
            The Orcale Identity Service Client.
        """
        self.ds_client = ds_client
        self.identity_client = identity_client
        self.user_name = ""
        self._etag = model_etag
        self._provenance_metadata_etag = provenance_etag
        self.provenance_metadata = provenance_metadata
        self._extract_oci_model(model)
        self._extract_user_name(model)

    def _extract_oci_model(self, model: OCIModel) -> None:
        """Extracts the model information from OCI model."""
        for key in model.swagger_types.keys():
            if key not in self._NEW_ATTRIBUTES:
                val = getattr(model, key)
                setattr(self, key, val)
        self.schema_input = self._extract_schema("input_schema", model)
        self.schema_output = self._extract_schema("output_schema", model)
        self.metadata_custom = self._extract_metadata_custom(model)
        self.metadata_taxonomy = self._extract_metadata_taxonomy(model)
        self.swagger_types = model.swagger_types
        self.lifecycle_state = model.lifecycle_state

    def _validate_metadata(self):
        self.metadata_custom.validate()
        self.metadata_taxonomy.validate()
        total_size = self.metadata_custom.size() + self.metadata_taxonomy.size()
        if total_size > METADATA_SIZE_LIMIT:
            raise MetadataSizeTooLarge(total_size)
        return True

    def _extract_user_name(self, model: OCIModel) -> None:
        try:
            user = self.identity_client.get_user(model.created_by)
            self.user_name = user.data.name
        except:
            pass

    @staticmethod
    def _extract_schema(key, model):
        """Extracts the input and output schema."""
        schema = Schema()
        if hasattr(model, key):
            try:
                schema = (
                    Schema.from_dict(json.loads(getattr(model, key)))
                    if getattr(model, key)
                    else Schema()
                )
            except Exception as e:
                logger.warning(str(e))
        return schema

    @staticmethod
    def _extract_metadata_taxonomy(model):
        """Extracts the taxonomy metadata."""
        metadata_taxonomy = ModelTaxonomyMetadata()
        if hasattr(model, "defined_metadata_list"):
            try:
                metadata_taxonomy = ModelTaxonomyMetadata._from_oci_metadata(
                    model.defined_metadata_list
                )
            except Exception as e:
                logger.warning(str(e))
        return metadata_taxonomy

    @staticmethod
    def _extract_metadata_custom(model):
        """Extracts the custom metadata."""
        metadata_custom = ModelCustomMetadata()
        if hasattr(model, "custom_metadata_list"):
            try:
                metadata_custom = ModelCustomMetadata._from_oci_metadata(
                    model.custom_metadata_list
                )
            except Exception as e:
                logger.warning(str(e))
        return metadata_custom

    def _to_dict(self):
        """Converts the model attributes to dictionary format."""
        attributes = {}
        for key in _UPDATE_MODEL_DETAILS_ATTRIBUTES:
            if hasattr(self, key):
                attributes[key] = getattr(self, key)

        if self.provenance_metadata is not None:
            attributes.update(
                {
                    key: getattr(self.provenance_metadata, key)
                    for key in _MODEL_PROVENANCE_ATTRIBUTES
                }
            )
        for field in self._FIELDS_TO_DECORATE:
            attributes[field] = getattr(self, field).to_dict()
        return attributes

    def _to_yaml(self):
        """Converts the model attributes to yaml format."""
        attributes = self._to_dict()
        return yaml.safe_dump(attributes)

    def to_dataframe(self) -> pd.DataFrame:
        """
        Converts the model to dataframe format.

        Returns
        -------
        panadas.DataFrame
            Pandas dataframe.
        """
        attributes = self._to_dict()
        df = pd.DataFrame.from_dict(attributes, orient="index", columns=[""]).dropna()
        return df

    @runtime_dependency(module="IPython", install_from=OptionalDependency.NOTEBOOK)
    def show_in_notebook(self, display_format: str = "dataframe") -> None:
        """Shows model in dataframe or yaml format.
        Supported formats: `dataframe` and `yaml`. Defaults to dataframe format.

        Returns
        -------
        None
            Nothing.
        """
        if display_format == "dataframe":

            from IPython.core.display import display

            display(self.to_dataframe())
        elif display_format == "yaml":
            print(self._to_yaml())
        else:
            NotImplementedError(
                "`display_format` is not supported. Choose 'dataframe' or 'yaml'"
            )

    def _repr_html_(self):
        """Shows model in dataframe format."""
        return (
            self.to_dataframe().style.set_properties(**{"margin-left": "0px"}).render()
        )

    def __repr__(self):
        """Shows model in dataframe format."""
        return (
            self.to_dataframe().style.set_properties(**{"margin-left": "0px"}).render()
        )

    def activate(self) -> None:
        """Activates model.

        Returns
        -------
        None
            Nothing.
        """
        self.lifecycle_state = OCIModel.LIFECYCLE_STATE_ACTIVE

    def deactivate(self) -> None:
        """Deactivates model.

        Returns
        -------
        None
            Nothing.
        """
        self.lifecycle_state = OCIModel.LIFECYCLE_STATE_INACTIVE

    def commit(self, force: bool = True) -> None:
        """Commits model changes.

        Parameters
        ----------
        force: bool
            If True, any remote changes on this model would be lost.

        Returns
        -------
        None
            Nothing.
        """
        self._validate_metadata()
        attributes = {
            key: getattr(self, key) for key in _UPDATE_MODEL_DETAILS_ATTRIBUTES
        }

        if hasattr(self, "metadata_custom"):
            attributes["custom_metadata_list"] = self.metadata_custom._to_oci_metadata()
        if hasattr(self, "metadata_taxonomy"):
            attributes[
                "defined_metadata_list"
            ] = self.metadata_taxonomy._to_oci_metadata()

        update_model_details = UpdateModelDetails(**attributes)
        # freeform_tags=self._model.freeform_tags, defined_tags=self._model.defined_tags)

        # update model
        # https://docs.oracle.com/en-us/iaas/Content/API/Concepts/usingapi.htm#eleven
        # The API supports etags for the purposes of optimistic concurrency control.
        # The GET and POST calls return an etag response header with a value you should store.
        # When you later want to update or delete the resource, set the if-match header to the ETag
        # you received for the resource. The resource will then be updated or deleted
        # only if the ETag you provide matches the current value of that resource's ETag.
        kwargs = {}
        if not force:
            kwargs["if_match"] = self._etag

        self.ds_client.update_model(
            self.id, update_model_details=update_model_details, **kwargs
        )
        # store the lifecycle status, as updating the model will delete info not included in "update_model_details"
        lifecycle_status = self.lifecycle_state
        self.__dict__.update(self._load_model().__dict__)
        self.lifecycle_state = lifecycle_status

        # update model state
        if not force:
            kwargs["if_match"] = self._etag
        if self.lifecycle_state == OCIModel.LIFECYCLE_STATE_ACTIVE:
            self.ds_client.activate_model(self.id, **kwargs)
        elif self.lifecycle_state == OCIModel.LIFECYCLE_STATE_INACTIVE:
            self.ds_client.deactivate_model(self.id, **kwargs)
        self.__dict__.update(self._load_model().__dict__)

        # update model provenance
        if self.provenance_metadata != ModelProvenance():
            if not force:
                kwargs["if_match"] = self._provenance_metadata_etag
            response = self.ds_client.update_model_provenance(
                self.id, self.provenance_metadata, **kwargs
            )
        # get model etag again, as updating model provenance changes it
        self.__dict__.update(self._load_model().__dict__)

    @staticmethod
    def _get_provenance_metadata(ds_client: DataScienceClient, model_id: str):
        """Gets provenance information for specified model."""
        try:
            provenance_response = ds_client.get_model_provenance(model_id)
        except ServiceError as e:
            if e.status == 404:
                try:
                    provenance_response = ds_client.create_model_provenance(
                        model_id, ModelProvenance()
                    )
                except ServiceError as e2:
                    raise e2
            elif e.status == 409:
                print("The model has been deleted.")
                raise e
            else:
                raise e
        return provenance_response

    @classmethod
    def load_model(
        cls,
        ds_client: DataScienceClient,
        identity_client: IdentityClient,
        model_id: str,
    ) -> "Model":
        """Loads the model from the model catalog based on model ID.

        Parameters
        ----------
        ds_client: DataScienceClient
            The Oracle DataScience client.
        identity_client: IdentityClient
            The Orcale Identity Service Client.
        model_id: str
            The model ID.

        Returns
        -------
        Model
            The ADS model catalog item.

        Raises
        ------
        ServiceError: If error occures while getting model from server.
        KeyError: If model not found.
        ValueError: If error occures while getting model provenance mettadata from server.
        """

        try:
            model_response = ds_client.get_model(model_id)
        except ServiceError as e:
            if e.status == 404:
                raise KeyError(e.message) from e
            raise e

        try:
            provenance_response = cls._get_provenance_metadata(ds_client, model_id)
        except Exception as e:
            raise ValueError(
                f"Unable to fetch model provenance metadata for model {model_id}"
            )

        return cls(
            model_response.data,
            _get_etag(model_response),
            provenance_response.data,
            _get_etag(provenance_response),
            ds_client,
            identity_client,
        )

    def _load_model(self):
        """Loads the model from model catalog."""
        return self.load_model(self.ds_client, self.identity_client, self.id)

    def rollback(self) -> None:
        """Rollbacks the changes made to the model.

        Returns
        -------
        None
            Nothing.
        """
        self.__dict__.update(self._load_model().__dict__)


class ModelCatalog:
    """
    Allows to list, load, update, download, upload and delete models from model catalog.

    Methods
    -------
    get_model(self, model_id)
        Loads the model from the model catalog based on model_id.
    list_models(self, project_id=None, include_deleted=False, datetime_format=utils.date_format, **kwargs)
        Lists all models in a given compartment, or in the current project if project_id is specified.
    list_model_deployment(self, model_id, config=None, tenant_id=None, limit=500, page=None, **kwargs)
        Gets the list of model deployments by model Id across the compartments.
    update_model(self, model_id, update_model_details=None, **kwargs)
        Updates a model with given model_id, using the provided update data.
    delete_model(self, model, **kwargs)
        Deletes the model based on model_id.
    download_model(self, model_id, target_dir, force_overwrite=False, install_libs=False, conflict_strategy=ConflictStrategy.IGNORE)
        Downloads the model from model_dir to target_dir based on model_id.
    upload_model(self, model_artifact, provenance_metadata=None, project_id=None, display_name=None, description=None)
        Uploads the model artifact to cloud storage.
    """

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def __init__(
        self,
        compartment_id: Optional[str] = None,
        ds_client_auth: Optional[dict] = None,
        identity_client_auth: Optional[dict] = None,
        timeout: Optional[int] = None,
        ds_client: Optional[DataScienceClient] = None,
        identity_client: Optional[IdentityClient] = None,
    ):
        """Initializes model catalog instance.

        Parameters
        ----------
        compartment_id : (str, optional). Defaults to None.
            Model compartment OCID. If `None`, the `config.NB_SESSION_COMPARTMENT_OCID` would be used.
        ds_client_auth : (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate DataScienceClient object.
        identity_client_auth : (dict, optional). Defaults to None.
            The default authetication is set using `ads.set_auth` API. If you need to override the
            default, use the `ads.common.auth.api_keys` or `ads.common.auth.resource_principal` to create appropriate
            authentication signer and kwargs required to instantiate IdentityClient object.
        timeout: (int, optional). Defaults to 10 seconds.
            The connection timeout in seconds for the client.
        ds_client: DataScienceClient
            The Oracle DataScience client.
        identity_client: IdentityClient
            The Orcale Identity Service Client.

        Raises
        ------
        ValueError
            If compartment ID not specified.
        TypeError
            If timeout not an integer.
        """
        self.compartment_id = (
            NB_SESSION_COMPARTMENT_OCID if compartment_id is None else compartment_id
        )
        if self.compartment_id is None:
            raise ValueError("compartment_id needs to be specified.")

        if timeout and not isinstance(timeout, int):
            raise TypeError("Timeout must be an integer.")

        self.ds_client_auth = ds_client_auth
        self.identity_client_auth = identity_client_auth
        self.ds_client = ds_client
        self.identity_client = identity_client

        if not self.ds_client:
            self.ds_client_auth = (
                ds_client_auth
                if ds_client_auth
                else auth.default_signer(
                    {"service_endpoint": OCI_ODSC_SERVICE_ENDPOINT}
                )
            )
            if timeout:
                if not self.ds_client_auth.get("client_kwargs"):
                    self.ds_client_auth["client_kwargs"] = {}
                self.ds_client_auth["client_kwargs"]["timeout"] = timeout
            self.ds_client = oci_client.OCIClientFactory(
                **self.ds_client_auth
            ).data_science

        if not self.identity_client:
            self.identity_client_auth = (
                identity_client_auth
                if identity_client_auth
                else auth.default_signer(
                    {"service_endpoint": OCI_ODSC_SERVICE_ENDPOINT}
                )
            )
            if timeout:
                if not self.identity_client_auth.get("client_kwargs"):
                    self.identity_client_auth["client_kwargs"] = {}
                self.identity_client_auth["client_kwargs"]["timeout"] = timeout
            self.identity_client = oci_client.OCIClientFactory(
                **self.identity_client_auth
            ).identity

        self.short_id_index = {}

    def __getitem__(self, model_id):  # pragma: no cover
        return self.get_model(model_id)

    def __contains__(self, model_id):  # pragma: no cover
        try:
            return self.get_model(model_id) is not None
        except KeyError:
            return False
        except Exception:
            raise

    def __iter__(self):  # pragma: no cover
        return self.list_models().__iter__()

    def __len__(self):  # pragma: no cover
        return len(self.list_models())

    def get_model(self, model_id):
        """
        Loads the model from the model catalog based on model_id.

        Parameters
        ----------
        model_id: str, required
            The model ID.

        Returns
        -------
        ads.catalog.Model
            The ads.catalog.Model with the matching ID.
        """
        if not model_id.startswith("ocid"):
            model_id = self.short_id_index[model_id]
            self.id = model_id
        return Model.load_model(self.ds_client, self.identity_client, model_id)

    def list_models(
        self,
        project_id: str = None,
        include_deleted: bool = False,
        datetime_format: str = utils.date_format,
        **kwargs,
    ):
        """
        Lists all models in a given compartment, or in the current project if project_id is specified.

        Parameters
        ----------
        project_id: str
            The project_id of model.
        include_deleted: bool, optional, default=False
            Whether to include deleted models in the returned list.
        datetime_format: str, optional, default: '%Y-%m-%d %H:%M:%S'
            Change format for date time fields.

        Returns
        -------
        ModelSummaryList
            A list of models.
        """
        try:
            list_models_response = self.ds_client.list_models(
                self.compartment_id, project_id=project_id, **kwargs
            )
            if list_models_response.data is None or len(list_models_response.data) == 0:
                print("No model found.")
                return
        except ServiceError as se:
            if se.status == 404:
                raise KeyError(se.message) from se
            else:
                raise

        model_list_filtered = [
            Model(
                model=model,
                model_etag=None,
                provenance_metadata=None,
                provenance_etag=None,
                ds_client=self.ds_client,
                identity_client=self.identity_client,
            )
            for model in list_models_response.data
            if include_deleted
            or model.lifecycle_state != ModelSummary.LIFECYCLE_STATE_DELETED
        ]
        # handle empty list
        if model_list_filtered is None or len(model_list_filtered) == 0:
            print("No model found.")
            return []

        msl = ModelSummaryList(
            self,
            model_list_filtered,
            list_models_response,
            datetime_format=datetime_format,
        )
        self.short_id_index.update(msl.short_id_index)
        return msl

    def list_model_deployment(
        self,
        model_id: str,
        config: dict = None,
        tenant_id: str = None,
        limit: int = 500,
        page: str = None,
        **kwargs,
    ):
        """
        Gets the list of model deployments by model Id across the compartments.

        Parameters
        ----------
        model_id: str
            The model ID.
        config: dict (optional)
            Configuration keys and values as per SDK and Tool Configuration.
            The from_file() method can be used to load configuration from a file.
            Alternatively, a dict can be passed. You can validate_config the dict
            using validate_config(). Defaults to None.
        tenant_id: str (optional)
            The tenancy ID, which can be used to specify a different tenancy
            (for cross-tenancy authorization) when searching for resources in
            a different tenancy. Defaults to None.
        limit: int (optional)
            The maximum number of items to return. The value must be between
            1 and 1000. Defaults to 500.
        page: str (optional)
            The page at which to start retrieving results.

        Returns
        -------
            The list of model deployments.
        """
        query = f"query datasciencemodeldeployment resources where ModelId='{model_id}'"
        return OCIResource.search(
            query,
            type=SEARCH_TYPE.STRUCTURED,
            config=config,
            tenant_id=tenant_id,
            limit=limit,
            page=page,
            **kwargs,
        )

    def update_model(self, model_id, update_model_details=None, **kwargs) -> Model:
        """
        Updates a model with given model_id, using the provided update data.

        Parameters
        ----------
        model_id: str
            The model ID.
        update_model_details: UpdateModelDetails
            Contains the update model details data to apply.
            Mandatory unless kwargs are supplied.
        kwargs: dict, optional
            Update model details can be supplied instead as kwargs.

        Returns
        -------
        Model
            The ads.catalog.Model with the matching ID.
        """
        if not model_id.startswith("ocid"):
            model_id = self.short_id_index[model_id]
        if update_model_details is None:
            update_model_details = UpdateModelDetails(
                **{
                    k: v
                    for k, v in kwargs.items()
                    if k in _UPDATE_MODEL_DETAILS_ATTRIBUTES
                }
            )
            update_model_details.compartment_id = self.compartment_id
            # filter kwargs removing used keys
            kwargs = {
                k: v
                for k, v in kwargs.items()
                if k not in _UPDATE_MODEL_DETAILS_ATTRIBUTES
            }
        update_model_response = self.ds_client.update_model(
            model_id, update_model_details, **kwargs
        )
        provenance_response = Model._get_provenance_metadata(self.ds_client, model_id)
        return Model(
            model=update_model_response.data,
            model_etag=_get_etag(update_model_response),
            provenance_metadata=provenance_response.data,
            provenance_etag=_get_etag(provenance_response),
            ds_client=self.ds_client,
            identity_client=self.identity_client,
        )

    def delete_model(self, model: Union[str, "ads.catalog.Model"], **kwargs) -> bool:
        """
        Deletes the model from Model Catalog.

        Parameters
        ----------
        model: Union[str, "ads.catalog.Model"]
            The OCID of the model to delete as a string, or a `ads.catalog.Model` instance.

        kwargs:
            delete_associated_model_deployment: (bool, optional). Defaults to `False`.
                Whether associated model deployments need to be deletet or not.

        Returns
        -------
        bool
            `True` if the model was successfully deleted.

        Raises
        ------
        ModelWithActiveDeploymentError
            If model has active model deployments ant inout attribute
            `delete_associated_model_deployment` set to `False`.
        """
        model_id = (
            model.id
            if isinstance(model, Model)
            else self.short_id_index[model]
            if not model.startswith("ocid")
            else model
        )
        delete_associated_model_deployment = kwargs.pop(
            "delete_associated_model_deployment", None
        )
        active_deployments = tuple(
            item
            for item in self.list_model_deployment(model_id)
            if item.lifecycle_state == "ACTIVE"
        )

        if len(active_deployments) > 0:
            if not delete_associated_model_deployment:
                raise ModelWithActiveDeploymentError(
                    f"The model `{model_id}` has active model deployments: "
                    f"{[item.identifier for item in active_deployments]}. "
                    "Delete associated model deployments before deleting the model or "
                    "set the `delete_associated_model_deployment` attribute to `True`."
                )

            logger.info(
                f"Deleting model deployments associated with the model `{model_id}`."
            )
            for oci_model_deployment in active_deployments:
                (
                    ModelDeployer(config=self.ds_client_auth)
                    .get_model_deployment(oci_model_deployment.identifier)
                    .delete(wait_for_completion=True)
                )

        logger.info(f"Deleting model `{model_id}`.")
        self.ds_client.delete_model(model_id, **kwargs)
        return True

    def _download_artifact(
        self,
        model_id: str,
        target_dir: str,
        force_overwrite: Optional[bool] = False,
        bucket_uri: Optional[str] = None,
        remove_existing_artifact: Optional[bool] = True,
    ) -> None:
        """
        Downloads the model artifacts from model catalog to target_dir based on `model_id`.

        Parameters
        ----------
        model_id: str
            The OCID of the model to download.
        target_dir: str
            The target location of model artifacts.
        force_overwrite: (bool, optional). Defaults to `False`.
            Overwrite target directory if exists.
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for downloading large artifacts with
            size is greater than 2GB. Example: `bucket_uri=oci://<bucket_name>@<namespace>/prefix/`.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Whether artifacts uploaded to object storage bucket need to be removed or not.

        Raises
        ------
        ValueError
            If targeted directory does not exist.
        KeyError
            If model id not found.

        Returns
        -------
        None
            Nothing
        """
        if os.path.exists(target_dir) and os.listdir(target_dir):
            if not force_overwrite:
                raise ValueError(
                    f"The `{target_dir}` directory already exists. "
                    "Set `force_overwrite` to `True` if you wish to overwrite."
                )
            shutil.rmtree(target_dir)

        with utils.get_progress_bar(6) as progress:
            progress.update("Getting information about model artifacts")

            # If the size of artifacts greater than 2GB, then artifacts
            # need to be imported to OS bucket at first.
            try:
                model_artifact_info = self.ds_client.head_model_artifact(
                    model_id=model_id
                ).headers
            except ServiceError as ex:
                if ex.status == 404:
                    raise KeyError(f"The model `{model_id}` not found.") from ex
                raise

            artifact_size = int(model_artifact_info.get("content-length"))

            if bucket_uri or artifact_size > _MAX_ARTIFACT_SIZE_IN_BYTES:
                if not bucket_uri:
                    raise ModelArtifactSizeError(
                        utils.human_size(_MAX_ARTIFACT_SIZE_IN_BYTES)
                    )

                self._download_large_artifact(
                    model_id=model_id,
                    target_dir=target_dir,
                    bucket_uri=os.path.join(bucket_uri, f"{model_id}.zip"),
                    progress=progress,
                    remove_existing_artifact=remove_existing_artifact,
                )
            else:
                self._download_small_artifact(
                    model_id=model_id, target_dir=target_dir, progress=progress
                )
                progress.update()

            progress.update("Done")

    def _download_large_artifact(
        self,
        model_id: str,
        target_dir: str,
        bucket_uri: str,
        progress: TqdmProgressBar,
        remove_existing_artifact: Optional[bool] = True,
    ) -> None:
        """
        Downloads the model artifacts from model catalog to target_dir based on `model_id`.
        This method is used for artifacts with size greater than 2GB.

        Parameters
        ----------
        model_id: str
            The OCID of the model to download.
        target_dir: str
            The target location of model artifacts.
        bucket_uri: str
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for downloading large artifacts with
            size is greater than 2GB. Example: `bucket_uri=oci://<bucket_name>@<namespace>/prefix/`.
        progress: TqdmProgressBar
            The progress bar.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Whether artifacts uploaded to object storage bucket need to be removed or not.

        Returns
        -------
        None
            Nothing.
        """
        progress.update(f"Importing model artifacts from model catalog")
        self._import_model_artifact(model_id=model_id, bucket_uri=bucket_uri)

        progress.update("Copying model artifacts to the artifact directory")
        with tempfile.TemporaryDirectory() as temp_dir:
            zip_file_path = os.path.join(temp_dir, f"{str(uuid.uuid4())}.zip")
            zip_file_path = utils.copy_file(
                uri_src=bucket_uri,
                uri_dst=zip_file_path,
                progressbar_description="Copying model artifacts to the artifact directory",
            )

            progress.update("Extracting model artifacts")
            with ZipFile(zip_file_path) as zip_file:
                zip_file.extractall(target_dir)

        if remove_existing_artifact:
            progress.update("Removing temporary artifacts")
            utils.remove_file(bucket_uri, self.ds_client_auth)
        else:
            progress.update()

    def _import_model_artifact(
        self,
        model_id: str,
        bucket_uri: str,
    ):
        """Imports model artifact from the model catalog to the object storage bucket.
        This method is used for the case when the artifact size is greater than 2GB.
        """
        bucket_details = ObjectStorageDetails.from_path(bucket_uri)
        response = self.ds_client.import_model_artifact(
            model_id=model_id,
            import_model_artifact_details=ImportModelArtifactDetails(
                artifact_import_details=ArtifactImportDetailsObjectStorage(
                    namespace=bucket_details.namespace,
                    destination_bucket=bucket_details.bucket,
                    destination_object_name=bucket_details.filepath,
                    destination_region=self._region,
                )
            ),
        )

        self._wait_for_work_request(
            response=response,
            first_step_description="Preparing to import model artifacts from the model catalog",
            num_steps=3,
        )

    def _download_small_artifact(
        self,
        model_id: str,
        target_dir: str,
        progress: TqdmProgressBar,
    ) -> None:
        """
        Downloads the model artifacts from model catalog to target_dir based on `model_id`.
        This method is used for the artifacts with size less than 2GB.

        Parameters
        ----------
        model_id: str
            The OCID of the model to download.
        target_dir: str
            The target location of model artifacts.
        progress: TqdmProgressBar
            The progress bar.

        Returns
        -------
        None
            Nothing
        """
        progress.update("Importing model artifacts from catalog")
        try:
            zip_contents = self.ds_client.get_model_artifact_content(
                model_id
            ).data.content
        except ServiceError as ex:
            if ex.status == 404:
                raise KeyError(ex.message) from ex
            raise

        with tempfile.TemporaryDirectory() as temp_dir:
            progress.update("Copying model artifacts to the artifact directory")
            zip_file_path = os.path.join(temp_dir, f"{str(uuid.uuid4())}.zip")
            with open(zip_file_path, "wb") as zip_file:
                zip_file.write(zip_contents)
            progress.update("Extracting model artifacts")
            with ZipFile(zip_file_path) as zip_file:
                zip_file.extractall(target_dir)

    @deprecated(
        "2.5.9",
        details="Instead use `ads.common.model_artifact.ModelArtifact.from_model_catalog()`.",
    )
    def download_model(
        self,
        model_id: str,
        target_dir: str,
        force_overwrite: bool = False,
        install_libs: bool = False,
        conflict_strategy=ConflictStrategy.IGNORE,
        bucket_uri: Optional[str] = None,
        remove_existing_artifact: Optional[bool] = True,
    ):
        """
        Downloads the model from model_dir to target_dir based on model_id.

        Parameters
        ----------
        model_id: str
            The OCID of the model to download.
        target_dir: str
            The target location of model after download.
        force_overwrite: bool
            Overwrite target_dir if exists.
        install_libs: bool, default: False
            Install the libraries specified in ds-requirements.txt which are missing in the current environment.
        conflict_strategy: ConflictStrategy, default: IGNORE
           Determines how to handle version conflicts between the current environment and requirements of
           model artifact.
           Valid values: "IGNORE", "UPDATE" or ConflictStrategy.
           IGNORE: Use the installed version in  case of conflict
           UPDATE: Force update dependency to the version required by model artifact in case of conflict
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for downloading large artifacts with
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Whether artifacts uploaded to object storage bucket need to be removed or not.

        Returns
        -------
        ModelArtifact
            A ModelArtifact instance.
        """
        self._download_artifact(
            model_id,
            target_dir,
            force_overwrite,
            bucket_uri=bucket_uri,
            remove_existing_artifact=remove_existing_artifact,
        )

        result = ModelArtifact(
            target_dir,
            conflict_strategy=conflict_strategy,
            install_libs=install_libs,
            reload=False,
        )

        try:
            model_response = self.ds_client.get_model(model_id)
        except ServiceError as e:
            if e.status == 404:
                raise KeyError(e.message) from e
            raise e

        if hasattr(model_response.data, "custom_metadata_list"):
            try:
                result.metadata_custom = ModelCustomMetadata._from_oci_metadata(
                    model_response.data.custom_metadata_list
                )
            except:
                result.metadata_custom = ModelCustomMetadata()
        if hasattr(model_response.data, "defined_metadata_list"):
            try:
                result.metadata_taxonomy = ModelTaxonomyMetadata._from_oci_metadata(
                    model_response.data.defined_metadata_list
                )
            except:
                result.metadata_taxonomy = ModelTaxonomyMetadata()
        if hasattr(model_response.data, "input_schema"):
            try:
                result.schema_input = Schema.from_dict(
                    json.loads(model_response.data.input_schema)
                    if model_response.data.input_schema != ""
                    else Schema()
                )
            except:
                result.schema_input = Schema()
        if hasattr(model_response.data, "output_schema"):
            try:
                result.schema_output = Schema.from_dict(
                    json.loads(model_response.data.output_schema)
                    if model_response.data.output_schema != ""
                    else Schema()
                )
            except:
                result.schema_output = Schema()

        if not install_libs:
            logger.warning(
                "Libraries in `ds-requirements.txt` were not installed. "
                "Use `install_requirements()` to install the required dependencies."
            )
        return result

    @deprecated(
        "2.6.6",
        details="Use framework specific Model utility class for saving and deploying model. Check https://accelerated-data-science.readthedocs.io/en/latest/user_guide/model_registration/quick_start.html",
    )
    def upload_model(
        self,
        model_artifact: ModelArtifact,
        provenance_metadata: Optional[ModelProvenance] = None,
        project_id: Optional[str] = None,
        display_name: Optional[str] = None,
        description: Optional[str] = None,
        freeform_tags: Optional[Dict[str, Dict[str, object]]] = None,
        defined_tags: Optional[Dict[str, Dict[str, object]]] = None,
        bucket_uri: Optional[str] = None,
        remove_existing_artifact: Optional[bool] = True,
        overwrite_existing_artifact: Optional[bool] = True,
        model_version_set: Optional[Union[str, ModelVersionSet]] = None,
        version_label: Optional[str] = None,
    ):
        """
        Uploads the model artifact to cloud storage.

        Parameters
        ----------
        model_artifact: Union[ModelArtifact, GenericModel]
            The model artifacts or generic model instance.
        provenance_metadata: (ModelProvenance, optional). Defaults to None.
            Model provenance gives data scientists information about the origin of their model.
            This information allows data scientists to reproduce
            the development environment in which the model was trained.
        project_id: (str, optional). Defaults to None.
            The project_id of model.
        display_name: (str, optional). Defaults to None.
            The name of model. If a display_name is not provided, a randomly generated easy to remember name
            with timestamp will be generated, like 'strange-spider-2022-08-17-23:55.02'.
        description: (str, optional). Defaults to None.
            The description of model.
        freeform_tags: (Dict[str, str], optional). Defaults to None.
            Freeform tags for the model, by default None
        defined_tags: (Dict[str, dict[str, object]], optional). Defaults to None.
            Defined tags for the model, by default None.
        bucket_uri: (str, optional). Defaults to None.
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for uploading large artifacts which
            size greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Whether artifacts uploaded to object storage bucket need to be removed or not.
        overwrite_existing_artifact: (bool, optional). Defaults to `True`.
            Overwrite target bucket artifact if exists.
        model_version_set: (Union[str, ModelVersionSet], optional). Defaults to None.
            The Model version set OCID, or name, or `ModelVersionSet` instance.
        version_label: (str, optional). Defaults to None.
            The model version label.

        Returns
        -------
        ads.catalog.Model
            The ads.catalog.Model with the matching ID.
        """
        project_id = project_id or PROJECT_OCID
        if not project_id:
            raise ValueError("`project_id` needs to be specified.")

        copy_artifact_to_os = False
        if (
            bucket_uri
            or utils.folder_size(model_artifact.artifact_dir)
            > _MAX_ARTIFACT_SIZE_IN_BYTES
        ):
            if not bucket_uri:
                raise ValueError(
                    f"The model artifacts size is greater than `{utils.human_size(_MAX_ARTIFACT_SIZE_IN_BYTES)}`. "
                    "The `bucket_uri` needs to be specified to "
                    "copy artifacts to the object storage bucket. "
                    "Example: `bucket_uri=oci://<bucket_name>@<namespace>/prefix/`"
                )
            copy_artifact_to_os = True

        # extract model_version_set_id from model_version_set attribute or environment
        # variables in case of saving model in context of model version set.
        model_version_set_id = _extract_model_version_set_id(model_version_set)
        # Set default display_name if not specified - randomly generated easy to remember name generated
        display_name = display_name or utils.get_random_name_for_resource()

        with utils.get_progress_bar(5) as progress:
            project_id = PROJECT_OCID if project_id is None else project_id
            if project_id is None:
                raise ValueError("project_id needs to be specified.")
            schema_file = os.path.join(model_artifact.artifact_dir, "schema.json")
            if os.path.exists(schema_file):
                with open(schema_file, "r") as schema:
                    metadata = json.load(schema)
                    freeform_tags = {"problem_type": metadata["problem_type"]}

            progress.update("Saving model in the model catalog")
            create_model_details = CreateModelDetails(
                display_name=display_name,
                description=description,
                project_id=project_id,
                compartment_id=self.compartment_id,
                custom_metadata_list=model_artifact.metadata_custom._to_oci_metadata()
                if model_artifact.metadata_custom is not None
                else [],
                defined_metadata_list=model_artifact.metadata_taxonomy._to_oci_metadata()
                if model_artifact.metadata_taxonomy is not None
                else [],
                input_schema=model_artifact.schema_input.to_json()
                if model_artifact.schema_input is not None
                else '{"schema": []}',
                output_schema=model_artifact.schema_output.to_json()
                if model_artifact.schema_output is not None
                else '{"schema": []}',
                freeform_tags=freeform_tags,
                defined_tags=defined_tags,
                model_version_set_id=model_version_set_id,
                version_label=version_label,
            )
            model = self.ds_client.create_model(create_model_details)

            if provenance_metadata is not None:
                progress.update("Saving provenance metadata")
                self.ds_client.create_model_provenance(
                    model.data.id, provenance_metadata
                )
            else:
                progress.update()

            # if the model artifact size greater than 2GB then export function
            # needs to be used instead of upload. The export function will copy
            # model artifacts to the OS bucket at first and then will upload
            # the artifacts to the model catalog.
            if copy_artifact_to_os:
                self._export_model_artifact(
                    model_id=model.data.id,
                    model_artifact=model_artifact,
                    progress=progress,
                    bucket_uri=bucket_uri,
                    remove_existing_artifact=remove_existing_artifact,
                    overwrite_existing_artifact=overwrite_existing_artifact,
                )
            else:
                self._upload_model_artifact(
                    model_id=model.data.id,
                    model_artifact=model_artifact,
                    progress=progress,
                )
                progress.update()

            progress.update("Done")
        return self.get_model(model.data.id)

    def _prepare_model_artifact(self, model_artifact, progress: TqdmProgressBar) -> str:
        """Prepares model artifacts to save in the Model Catalog.

        Returns
        -------
        str
            The path to the model artifact zip archive.
        """
        progress.update("Preparing model artifacts zip")
        files_to_upload = model_artifact._get_files()
        artifact_path = "/tmp/saved_model_" + str(uuid.uuid4()) + ".zip"
        zf = ZipFile(artifact_path, "w")
        for matched_file in files_to_upload:
            zf.write(
                os.path.join(model_artifact.artifact_dir, matched_file),
                arcname=matched_file,
            )
        zf.close()
        return artifact_path

    def _upload_model_artifact(self, model_id, model_artifact, progress):
        """Uploads model artifact to the model catalog.
        This method can be used only if the size of model artifact is less than 2GB.
        For the artifacts with size greater than 2 GB the `_export_model_artifact`
        method should be used instead.
        """
        artifact_zip_path = self._prepare_model_artifact(model_artifact, progress)
        progress.update("Uploading model artifacts to the catalog")
        with open(artifact_zip_path, "rb") as file_data:
            bytes_content = file_data.read()
            self.ds_client.create_model_artifact(
                model_id,
                bytes_content,
                content_disposition=f'attachment; filename="{model_id}.zip"',
            )
        os.remove(artifact_zip_path)
        progress.update()

    def _export_model_artifact(
        self,
        model_id: str,
        model_artifact: ModelArtifact,
        bucket_uri: str,
        progress,
        remove_existing_artifact: Optional[bool] = True,
        overwrite_existing_artifact: Optional[bool] = True,
    ):
        """Exports model artifact to the model catalog.
        This method is used for the case when the artifact size is greater than 2GB.
        1. Archive model artifact.
        2. Copies the artifact to the object storage bucket.
        3. Exports artifact from the user's object storage bucket to the system one.
        """
        artifact_zip_path = self._prepare_model_artifact(model_artifact, progress)
        progress.update(f"Copying model artifact to the Object Storage bucket")

        try:
            bucket_uri_file_name = os.path.basename(bucket_uri)
            if not bucket_uri_file_name:
                bucket_uri = os.path.join(bucket_uri, f"{model_id}.zip")
            elif not bucket_uri.lower().endswith(".zip"):
                bucket_uri = f"{bucket_uri}.zip"

            bucket_file_name = utils.copy_file(
                artifact_zip_path,
                bucket_uri,
                force_overwrite=overwrite_existing_artifact,
                auth=self.ds_client_auth,
                progressbar_description="Copying model artifact to the Object Storage bucket",
            )
        except FileExistsError:
            raise FileExistsError(
                f"The `{bucket_uri}` exists. Please use a new file name or "
                "set `overwrite_existing_artifact` to `True` if you wish to overwrite."
            )

        os.remove(artifact_zip_path)

        progress.update("Exporting model artifact to the model catalog")
        bucket_details = ObjectStorageDetails.from_path(bucket_file_name)
        response = self.ds_client.export_model_artifact(
            model_id=model_id,
            export_model_artifact_details=ExportModelArtifactDetails(
                artifact_export_details=ArtifactExportDetailsObjectStorage(
                    namespace=bucket_details.namespace,
                    source_bucket=bucket_details.bucket,
                    source_object_name=bucket_details.filepath,
                    source_region=self._region,
                )
            ),
        )

        self._wait_for_work_request(
            response=response,
            first_step_description="Preparing to export model artifact to the model catalog",
            num_steps=4,
        )

        if remove_existing_artifact:
            progress.update(
                "Removing temporary model artifact from the Object Storage bucket"
            )
            utils.remove_file(bucket_file_name, self.ds_client_auth)
        else:
            progress.update()

    @property
    def _region(self):
        """Gets current region."""
        if "region" in self.ds_client_auth.get("config", {}):
            return self.ds_client_auth["config"]["region"]
        return json.loads(OCI_REGION_METADATA)["regionIdentifier"]

    def _wait_for_work_request(
        self, response, first_step_description: str = "", num_steps=4
    ):
        """Waits for the work request to be completed."""
        STOP_STATE = (
            WorkRequest.STATUS_SUCCEEDED,
            WorkRequest.STATUS_CANCELED,
            WorkRequest.STATUS_CANCELING,
            WorkRequest.STATUS_FAILED,
        )
        work_request_id = response.headers["opc-work-request-id"]
        work_request_logs = None

        i = 0
        with utils.get_progress_bar(num_steps) as progress:
            progress.update(first_step_description)
            while not work_request_logs or len(work_request_logs) < num_steps:
                time.sleep(_WORK_REQUEST_INTERVAL_IN_SEC)
                work_request = self.ds_client.get_work_request(work_request_id)
                work_request_logs = self.ds_client.list_work_request_logs(
                    work_request_id
                ).data
                if work_request_logs:
                    new_work_request_logs = work_request_logs[i:]

                for wr_item in new_work_request_logs:
                    progress.update(wr_item.message)
                    i += 1

                if work_request.data.status in STOP_STATE:
                    if work_request.data.status != WorkRequest.STATUS_SUCCEEDED:
                        if work_request_logs:
                            raise Exception(work_request_logs[-1].message)
                        else:
                            raise Exception(
                                "An error occurred in attempt to perform the operation. Check the service logs to get more details."
                            )
                    else:
                        break
        return work_request



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2020, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import numpy as np
from ads.common.utils import _is_dask_dataframe, _is_dask_series


class ADSData(object):
    def __init__(self, X=None, y=None, name="", dataset_type=None):
        r"""
        This class wraps the input dataframe to various models, evaluation, and explanation frameworks.
        It's primary purpose is to hold any metadata relevant to these tasks. This can include it's:

        - X - the independent variables as some dataframe-like structure,
        - y - the dependent variable or target column as some array-like structure,
        - name - a string to name the data for user convenience,
        - dataset_type - the type of the X value.

        As part of this initiative, ADSData knows how to turn itself into an onnxruntime compatible data
        structure with the method .to_onnxrt(), which takes and onnx session as input.

        Parameters
        ----------
        X : Union[pandas.DataFrame, dask.DataFrame, numpy.ndarray, scipy.sparse.csr.csr_matrix]
            If str, URI for the dataset. The dataset could be read from local or network file system, hdfs, s3 and gcs
            Should be none if X_train, y_train, X_test, Y_test are provided
        y: Union[str, pandas.DataFrame, dask.DataFrame, pandas.Series, dask.Series, numpy.ndarray]
            If str, name of the target in X, otherwise series of labels corresponding to X
        name: str, optional
            Name to identify this data
        dataset_type: ADSDataset optional
            When this value is available, would be used to evaluate the ads task type
        kwargs:
            Additional keyword arguments that would be passed to the underlying Pandas read API.
        """
        self.X = X
        self.y = y
        self.name = name
        self.dataset_type = dataset_type

    @staticmethod
    def build(X=None, y=None, name="", dataset_type=None, **kwargs):
        r"""
        Returns an ADSData object built from the (source, target) or (X,y)

        Parameters
        ----------
        X : Union[pandas.DataFrame, dask.DataFrame, numpy.ndarray, scipy.sparse.csr.csr_matrix]
            If str, URI for the dataset. The dataset could be read from local or network file system, hdfs, s3 and gcs
            Should be none if X_train, y_train, X_test, Y_test are provided
        y: Union[str, pandas.DataFrame, dask.DataFrame, pandas.Series, dask.Series, numpy.ndarray]
            If str, name of the target in X, otherwise series of labels corresponding to X
        name: str, optional
            Name to identify this data
        dataset_type: ADSDataset, optional
            When this value is available, would be used to evaluate the ads task
            type
        kwargs:
            Additional keyword arguments that would be passed to the underlying Pandas read API.

        Returns
        -------
        ads_data: ads.common.data.ADSData
            A built ADSData object

        Examples
        --------
        >>> data = open_csv("my.csv")

        >>> data_ads = ADSData(data, 'target').build(data, 'target')
        """
        if X is None or y is None:
            raise ValueError("Both X and y are required.")
        if _is_dask_dataframe(X):
            X = X.compute()
        if _is_dask_series(y):
            y = y.compute()
        if dataset_type is None:
            dataset_type = type(X)
        if isinstance(y, str):
            try:
                return ADSData(
                    X.drop(y, axis=1), X[y], name=name, dataset_type=dataset_type
                )
            except AttributeError:
                raise ValueError(
                    "If y is a string, then X must be a pandas or dask dataframe"
                )
        else:
            return ADSData(X, y, name=name, dataset_type=dataset_type)

    def __repr__(self):
        return "%sShape of X:%s\nShape of y:%s" % (
            self.name + "\n",
            str(self.X.shape),
            str(self.y.shape),
        )

    def to_onnxrt(
        self, sess, idx_range=None, model=None, impute_values={}, **kwargs
    ):  # pragma: no cover
        r"""
        Returns itself formatted as an input for the onnxruntime session inputs passed in.

        Parameters
        ----------
        sess: Session
            The session object
        idx_range: Range
            The range of inputs to convert to onnx
        model: SupportedModel
            A model that supports being serialized for the onnx runtime.
        kwargs: additional keyword arguments

            - sess_inputs - Pass in the output from onnxruntime.InferenceSession("model.onnx").get_inputs()
            - input_dtypes (list) - If sess_inputs cannot be passed in, pass in the numpy dtypes of each input
            - input_shapes (list) - If sess_inputs cannot be passed in, pass in the shape of each input
            - input_names (list) -If sess_inputs cannot be passed in, pass in the name of each input

        Returns
        -------
        ort: Array
            array of inputs formatted for the given session.
        """
        if model._underlying_model in ["torch"]:
            sess_inputs = sess.get_inputs()
            in_shape, in_name, in_type = [], [], []
            for i, ftr in enumerate(sess_inputs):
                in_type.append(ftr.type)
                in_shape.append(ftr.shape)
                in_name.append(ftr.name)
            ret = {}
            for i, name in enumerate(in_name):
                idx_range = (0, len(self.X)) if idx_range is None else idx_range
                batch_size = idx_range[1] - idx_range[0]
                ret[name] = (
                    self.X[:batch_size]
                    .reshape([batch_size] + list(self.X[:1].shape))
                    .detach()
                    .cpu()
                    .numpy()
                    .astype(np.float32)
                )
            return ret
        elif model._underlying_model in ["automl"]:
            X_trans = model._onnx_data_transformer(
                X=self.X, impute_values=impute_values
            )
            inputs = {}
            for idx, c in enumerate(X_trans.columns):
                inputs[sess.get_inputs()[idx].name] = (
                    X_trans[c]
                    .values.reshape((X_trans.shape[0], 1))
                    .astype(X_trans.dtypes[idx])
                )
            return inputs
        elif model._underlying_model in ["lightgbm", "xgboost", "sklearn"]:
            idx_range = (0, len(self.X)) if idx_range is None else idx_range
            inputs = []
            for name, row in self.X[idx_range[0] : idx_range[1]].iterrows():
                inputs.append(list(row))
            return {"input": inputs}



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import os
import shutil
from abc import ABC, abstractmethod
from typing import Dict, Optional

from ads.common import utils
from ads.model.common import utils as model_utils
from ads.model.service.oci_datascience_model import OCIDataScienceModel


class ArtifactUploader(ABC):
    """The abstract class to upload model artifacts."""

    PROGRESS_STEPS_COUNT = 3

    def __init__(self, dsc_model: OCIDataScienceModel, artifact_path: str):
        """Initializes `ArtifactUploader` instance.

        Parameters
        ----------
        dsc_model: OCIDataScienceModel
            The data scince model instance.
        artifact_path: str
            The model artifact location.
        """
        if not os.path.exists(artifact_path):
            raise ValueError(f"The `{artifact_path}` does not exist")

        self.dsc_model = dsc_model
        self.artifact_path = artifact_path
        self.artifact_zip_path = None
        self.progress = None

    def upload(self):
        """Uploads model artifacts."""
        try:
            with utils.get_progress_bar(
                ArtifactUploader.PROGRESS_STEPS_COUNT + self.PROGRESS_STEPS_COUNT
            ) as progress:
                self.progress = progress
                self.progress.update("Preparing model artifacts ZIP archive.")
                self._prepare_artiact_tmp_zip()
                self.progress.update("Uploading model artifacts.")
                self._upload()
                self.progress.update(
                    "Uploading model artifacts has been successfully completed."
                )
                self.progress.update("Done.")
        except Exception:
            raise
        finally:
            self._remove_artiact_tmp_zip()

    def _prepare_artiact_tmp_zip(self) -> str:
        """Prepares model artifacts ZIP archive.

        Parameters
        ----------
        progress: (TqdmProgressBar, optional). Defaults to `None`.
            The progress indicator.

        Returns
        -------
        str
            Path to the model artifact ZIP archive.
        """
        if os.path.isfile(self.artifact_path) and self.artifact_path.lower().endswith(
            ".zip"
        ):
            self.artifact_zip_path = self.artifact_path
        else:
            self.artifact_zip_path = model_utils.zip_artifact(
                artifact_dir=self.artifact_path
            )
        return self.artifact_zip_path

    def _remove_artiact_tmp_zip(self):
        """Removes temporary created artifact zip archive."""
        if (
            self.artifact_zip_path
            and self.artifact_zip_path.lower() != self.artifact_path.lower()
        ):
            shutil.rmtree(self.artifact_zip_path, ignore_errors=True)

    @abstractmethod
    def _upload(self):
        """Uploads model artifacts. Needs to be implemented in a child class."""


class SmallArtifactUploader(ArtifactUploader):
    PROGRESS_STEPS_COUNT = 1

    def _upload(self):
        """Uploads model artifacts to the model catalog."""
        self.progress.update("Uploading model artifacts to the catalog")
        with open(self.artifact_zip_path, "rb") as file_data:
            self.dsc_model.create_model_artifact(file_data)


class LargeArtifactUploader(ArtifactUploader):
    PROGRESS_STEPS_COUNT = 4

    def __init__(
        self,
        dsc_model: OCIDataScienceModel,
        artifact_path: str,
        bucket_uri: str,
        auth: Optional[Dict] = None,
        region: Optional[str] = None,
        overwrite_existing_artifact: Optional[bool] = True,
        remove_existing_artifact: Optional[bool] = True,
    ):
        """Initializes `LargeArtifactUploader` instance.

        Parameters
        ----------
        dsc_model: OCIDataScienceModel
            The data scince model instance.
        artifact_path: str
            The model artifact location.
        bucket_uri: str
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for uploading large artifacts which
            size is greater than 2GB. Example: `oci://<bucket_name>@<namespace>/prefix/`.
        auth: (Dict, optional). Defaults to `None`.
            The default authetication is set using `ads.set_auth` API.
            If you need to override the default, use the `ads.common.auth.api_keys` or
            `ads.common.auth.resource_principal` to create appropriate authentication signer
            and kwargs required to instantiate IdentityClient object.
        region: (str, optional). Defaults to `None`.
            The destination Object Storage bucket region.
            By default the value will be extracted from the `OCI_REGION_METADATA` environment variables.
        overwrite_existing_artifact: (bool, optional). Defaults to `True`.
            Overwrite target bucket artifact if exists.
        remove_existing_artifact: (bool, optional). Defaults to `True`.
            Wether artifacts uploaded to object storage bucket need to be removed or not.
        """
        if not bucket_uri:
            raise ValueError("The `bucket_uri` must be provided.")

        super().__init__(dsc_model=dsc_model, artifact_path=artifact_path)
        self.auth = auth or dsc_model.auth
        self.region = region or utils.extract_region(self.auth)
        self.bucket_uri = bucket_uri
        self.overwrite_existing_artifact = overwrite_existing_artifact
        self.remove_existing_artifact = remove_existing_artifact

    def _upload(self):
        """Uploads model artifacts to the model catalog."""
        self.progress.update("Copying model artifact to the Object Storage bucket")

        try:
            bucket_uri = self.bucket_uri
            bucket_uri_file_name = os.path.basename(bucket_uri)

            if not bucket_uri_file_name:
                bucket_uri = os.path.join(bucket_uri, f"{self.dsc_model.id}.zip")
            elif not bucket_uri.lower().endswith(".zip"):
                bucket_uri = f"{bucket_uri}.zip"

            bucket_file_name = utils.copy_file(
                self.artifact_zip_path,
                bucket_uri,
                force_overwrite=self.overwrite_existing_artifact,
                auth=self.auth,
                progressbar_description="Copying model artifact to the Object Storage bucket",
            )
        except FileExistsError:
            raise FileExistsError(
                f"The `{self.bucket_uri}` exists. Please use a new file name or "
                "set `overwrite_existing_artifact` to `True` if you wish to overwrite."
            )
        self.progress.update("Exporting model artifact to the model catalog")
        self.dsc_model.export_model_artifact(
            bucket_uri=bucket_file_name, region=self.region
        )

        if self.remove_existing_artifact:
            self.progress.update(
                "Removing temporary artifacts from the Object Storage bucket"
            )
            utils.remove_file(bucket_uri)
        else:
            self.progress.update()



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from typing import Dict, List

from ads.common.oci_client import OCIClientFactory
from ads.common.auth import create_signer
from ads.opctl.utils import parse_conda_uri
import oci
import mmap
import json
import time
from tqdm import tqdm
from threading import Thread
import os


class MultiPartUploader:
    """
    Class that implements multipart uploading for conda packs.
    """

    def __init__(
        self,
        source_file: str,
        dst_uri: str,
        parts: int,
        oci_config: str = None,
        oci_profile: str = None,
        auth_type: str = None,
    ) -> None:
        """Initialize the class.

        Parameters
        ----------
        source_file : str
            path to conda pack file
        dst_uri : str
            path to destination of object storage location
        parts : int
            number of parts
        oci_config : str, optional
            path to oci config file, by default None
        oci_profile : str, optional
            oci profile to use, by default None
        auth_type : str
            authentication method, by default None
        """
        self.src = source_file
        self.dst = dst_uri
        self.oci_auth = create_signer(auth_type, oci_config, oci_profile)
        self.client = OCIClientFactory(**self.oci_auth).object_storage
        self.ns, self.bucket, self.path, _ = parse_conda_uri(dst_uri)
        self.file_size = os.path.getsize(self.src)
        # mmap offset arg must be a multiple of the ALLOCATIONGRANULARITY, change chunk_size to control offset value
        self.chunk_size = (
            (self.file_size // parts)
            // mmap.ALLOCATIONGRANULARITY
            * mmap.ALLOCATIONGRANULARITY
        )

    def upload(self, opc_meta: Dict = None) -> bool:
        """Uploading a conda pack to object storage.

        Parameters
        ----------
        opc_meta : Dict, optional
            metadata dictionary, by default None

        Returns
        -------
        bool
            whether uploading was successful

        Raises
        ------
        RuntimeError
            Uploading failed
        """
        multipart_upload_details = (
            oci.object_storage.models.CreateMultipartUploadDetails()
        )
        multipart_upload_details.object = self.path
        multipart_upload_details.metadata = {"opc-meta-manifest": json.dumps(opc_meta)}

        upload_details = self.client.create_multipart_upload(
            self.ns, self.bucket, multipart_upload_details
        )
        uploaded_parts = []
        threads = []
        mm_objects = []
        responses = []
        upload_id = upload_details.data.upload_id
        print(f"The upload id is {upload_id}.")
        with open(self.src, "rb") as pkf:
            counter = 0
            for offset, length in self._chunks():
                counter += 1
                mm = mmap.mmap(
                    pkf.fileno(), length, access=mmap.ACCESS_READ, offset=offset
                )
                mm_objects.append((mm, length))
                t = Thread(
                    target=self._upload_chunk,
                    args=(upload_id, counter, mm, uploaded_parts, responses),
                )
                t.start()
                threads.append(t)
            t = Thread(
                target=self._track_progress, args=(mm_objects, counter, responses)
            )
            t.start()
            threads.append(t)
            for t in threads:
                t.join()
        # Successful responses HTTP status in range (200–299)
        successful = all(200 <= r.status < 300 for r in responses)
        if successful:
            self.client.commit_multipart_upload(
                self.ns,
                self.bucket,
                self.path,
                upload_details.data.upload_id,
                oci.object_storage.models.CommitMultipartUploadDetails(
                    parts_to_commit=uploaded_parts
                ),
            )
            print(f"{self.src} uploaded successfuly to {self.dst}.")
        else:
            for r in responses:
                print(r.status, r.headers)
            raise RuntimeError(f"{self.src} upload failed.")
        return successful

    def _chunks(self):
        start_position = 0
        while start_position < self.file_size:
            yield start_position, min(self.chunk_size, self.file_size - start_position)
            start_position += self.chunk_size

    def _upload_chunk(
        self,
        upload_id: str,
        counter: int,
        mm: mmap.mmap,
        uploaded_parts: List,
        responses: List,
    ) -> None:
        response = self.client.upload_part(
            self.ns, self.bucket, self.path, upload_id, counter, mm
        )
        responses.append(response)
        uploaded_parts.append(
            oci.object_storage.models.CommitMultipartUploadPartDetails(
                etag=response.headers["etag"], part_num=counter
            )
        )

    @staticmethod
    def _track_progress(
        mm_objects: List[mmap.mmap], counter: int, responses: List
    ) -> None:
        nresponses = 0
        progress_objs = []
        for index, (mm_object, length) in enumerate(mm_objects):
            tqdm_obj = tqdm(
                total=length,
                unit="B",
                unit_scale=True,
                unit_divisor=2**10,
                desc=f"Part {index + 1}",
                position=index + 1,
                leave=True,
            )
            progress_objs.append([mm_object, tqdm_obj, 0, length])
        while nresponses < counter:
            for obj in progress_objs:
                current_pos = obj[0].tell()
                obj[1].update(current_pos - obj[2])
                obj[1].refresh()
                obj[2] = current_pos
            time.sleep(2)
            nresponses = len(responses)
        map(lambda x: x[1].refresh(), progress_objs)
        map(lambda x: x[1].close(), progress_objs)



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2022, 2023 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

import logging
import time
from functools import wraps
from io import BytesIO
from typing import Callable, Dict, List, Optional

import oci.data_science
from ads.common import utils
from ads.common.object_storage_details import ObjectStorageDetails
from ads.common.oci_datascience import OCIDataScienceMixin
from ads.common.oci_mixin import OCIWorkRequestMixin
from ads.common.oci_resource import SEARCH_TYPE, OCIResource
from ads.common.utils import extract_region
from ads.model.deployment.model_deployer import ModelDeployer
from oci.data_science.models import (
    ArtifactExportDetailsObjectStorage,
    ArtifactImportDetailsObjectStorage,
    CreateModelDetails,
    ExportModelArtifactDetails,
    ImportModelArtifactDetails,
    UpdateModelDetails,
    WorkRequest,
)
from oci.exceptions import ServiceError

logger = logging.getLogger(__name__)

_REQUEST_INTERVAL_IN_SEC = 3

MODEL_NEEDS_TO_BE_SAVED = (
    "Model needs to be saved to the Model Catalog before it can be accessed."
)


class ModelProvenanceNotFoundError(Exception):
    pass


class ModelArtifactNotFoundError(Exception):
    pass


class ModelNotSavedError(Exception):
    pass


class ModelWithActiveDeploymentError(Exception):
    pass


def check_for_model_id(msg: str = MODEL_NEEDS_TO_BE_SAVED):
    """The decorator helping to check if the ID attribute sepcified for a datascience model.

    Parameters
    ----------
    msg: str
        The message that will be thrown.

    Raises
    ------
    ModelNotSavedError
        In case if the ID attribute not specified.

    Examples
    --------
    >>> @check_for_id(msg="Some message.")
    ... def test_function(self, name: str, last_name: str)
    ...     pass
    """

    def decorator(func: Callable):
        @wraps(func)
        def wrapper(self, *args, **kwargs):
            if not self.id:
                raise ModelNotSavedError(msg)
            return func(self, *args, **kwargs)

        return wrapper

    return decorator


class OCIDataScienceModel(
    OCIDataScienceMixin,
    OCIWorkRequestMixin,
    oci.data_science.models.Model,
):
    """Represents an OCI Data Science Model.
    This class contains all attributes of the `oci.data_science.models.Model`.
    The main purpose of this class is to link the `oci.data_science.models.Model`
    and the related client methods.
    Linking the `Model` (payload) to Create/Update/Get/List/Delete methods.

    The `OCIDataScienceModel` can be initialized by unpacking the properties stored in a dictionary:

    .. code-block:: python

        properties = {
            "compartment_id": "<compartment_ocid>",
            "name": "<model_name>",
            "description": "<model_description>",
        }
        ds_model = OCIDataScienceModel(**properties)

    The properties can also be OCI REST API payload, in which the keys are in camel format.

    .. code-block:: python

        payload = {
            "compartmentId": "<compartment_ocid>",
            "name": "<model_name>",
            "description": "<model_description>",
        }
        ds_model = OCIDataScienceModel(**payload)

    Methods
    -------
    create(self) -> "OCIDataScienceModel"
        Creates datascience model in model catalog.
    create_model_provenance(self, model_provenance: ModelProvenance) -> oci.data_science.models.ModelProvenance:
        Creates model provenance metadata.
    def update_model_provenance(self, ModelProvenance) -> oci.data_science.models.ModelProvenance:
        Updates model provenance metadata.
    get_model_provenance(self) -> oci.data_science.models.ModelProvenance:
        Gets model provenance metadata.
    get_artifact_info(self) -> Dict:
        Gets model artifact attachment information.
    def get_model_artifact_content(self) -> BytesIO:
        Gets model artifact content.
    create_model_artifact(self, bytes_content: BytesIO) -> None:
        Creates model artifact for specified model.
    import_model_artifact(self, bucket_uri: str, region: str = None) -> None:
        Imports model artifact content from the model catalog.
    export_model_artifact(self, bucket_uri: str, region: str = None):
        Exports model artifact to the model catalog.
    update(self) -> "OCIDataScienceModel":
        Updates datascience Model.
    delete(self, delete_associated_model_deployment: Optional[bool] = False) -> "OCIDataScienceModel":
        Deletes detascience Model.
    model_deployment(self, ...) -> List:
        Gets the list of model deployments by model ID across the compartments.
    from_id(cls, ocid: str) -> "OCIDataScienceModel":
        Gets model by OCID.

    Examples
    --------
    >>> oci_model = OCIDataScienceModel.from_id(<model_ocid>)
    >>> oci_model.model_deployment()
    >>> oci_model.get_model_provenance()
    >>> oci_model.description = "A brand new description"
    ... oci_model.update()
    >>> oci_model.sync()
    >>> oci_model.get_artifact_info()
    """

    def create(self) -> "OCIDataScienceModel":
        """Creates datascience model in model catalog.

        Returns
        -------
        OCIDataScienceModel
            The `OCIDataScienceModel` instance (self), which allows chaining additional method.
        """
        if not self.compartment_id:
            raise ValueError("The `compartment_id` must be specified.")

        if not self.project_id:
            raise ValueError("The `project_id` must be specified.")

        return self.update_from_oci_model(
            self.client.create_model(self.to_oci_model(CreateModelDetails)).data
        )

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the provenance metadata can be created."
    )
    def create_model_provenance(
        self, model_provenance: oci.data_science.models.ModelProvenance
    ) -> oci.data_science.models.ModelProvenance:
        """Creates model provenance metadata.

        Parameters
        ----------
        model_provenance: oci.data_science.models.ModelProvenance
            OCI model provenance metadata.

        Returns
        -------
        oci.data_science.models.ModelProvenance
            The OCI model provenance object.
        """
        return self.client.create_model_provenance(self.id, model_provenance).data

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the provenance metadata can be updated."
    )
    def update_model_provenance(
        self, model_provenance: oci.data_science.models.ModelProvenance
    ) -> oci.data_science.models.ModelProvenance:
        """Updates model provenance metadata.

        Parameters
        ----------
        model_provenance: oci.data_science.models.ModelProvenance
            OCI model provenance metadata.

        Returns
        -------
        oci.data_science.models.ModelProvenance
            The OCI model provenance object.
        """
        return self.client.update_model_provenance(self.id, model_provenance).data

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the provenance metadata can be read."
    )
    def get_model_provenance(self) -> oci.data_science.models.ModelProvenance:
        """Gets model provenance metadata.

        Returns
        -------
        oci.data_science.models.ModelProvenance
            OCI model provenance metadata.

        Raises
        ------
        ModelProvenanceNotFoundError
            If model provenance not found.
        """
        try:
            return self.client.get_model_provenance(self.id).data
        except ServiceError as ex:
            if ex.status == 404:
                raise ModelProvenanceNotFoundError()
            raise

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the artifact information can be read."
    )
    def get_artifact_info(self) -> Dict:
        """Gets model artifact attachment information.

        Returns
        -------
        Dict
            The model artifact attachement info.
            Example:
            {
                'Date': 'Sun, 13 Nov 2022 06:01:27 GMT',
                'opc-request-id': 'E4F7',
                'ETag': '77156317-8bb9-4c4a-882b-0d85f8140d93',
                'Content-Disposition': 'attachment; filename=artifact.zip',
                'Last-Modified': 'Sun, 09 Oct 2022 16:50:14 GMT',
                'Content-Type': 'application/json',
                'Content-MD5': 'orMy3Gs386GZLjYWATJWuA==',
                'X-Content-Type-Options': 'nosniff',
                'Content-Length': '4029958'
            }

        Raises
        ------
        ModelArtifactNotFoundError
            If model artifact attchment not found.
        """
        try:
            return self.client.head_model_artifact(model_id=self.id).headers
        except ServiceError as ex:
            if ex.status == 404:
                raise ModelArtifactNotFoundError()
        return {}

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the artifact content can be read."
    )
    def get_model_artifact_content(self) -> BytesIO:
        """Gets model artifact content.
        Can only be used to the small artifacts, which size is less than 2GB.
        For the large artifacts needs to be used a `import_model_artifact` method.

        Returns
        -------
        BytesIO
            Object with data of type stream.

        Raises
        ------
        ModelArtifactNotFoundError
            If model artifact not found.

        """
        try:
            return self.client.get_model_artifact_content(model_id=self.id).data.content
        except ServiceError as ex:
            if ex.status == 404:
                raise ModelArtifactNotFoundError()

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the artifact can be created."
    )
    def create_model_artifact(self, bytes_content: BytesIO) -> None:
        """Creates model artifact for specified model.

        Parameters
        ----------
        bytes_content: BytesIO
            Model artifacts to upload.
        """
        self.client.create_model_artifact(
            self.id,
            bytes_content,
            content_disposition=f'attachment; filename="{self.id}.zip"',
        )

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the artifact can be created."
    )
    def import_model_artifact(self, bucket_uri: str, region: str = None) -> None:
        """Imports model artifact content from the model catalog.
        Requires to provide an Object Storage bucket for transitional saving artifacts.
        This method can be used either for small or large artifacts.

        Parameters
        ----------
        bucket_uri: str
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for downloading large artifacts which
            size is greater than 2GB.
            Example: `oci://<bucket_name>@<namespace>/prefix/`.
        region: (str, optional). Defaults to `None`.
            The destination Object Storage bucket region.
            By default the value will be extracted from the `OCI_REGION_METADATA` environment variable.

        Returns
        -------
        None

        Raises
        ------
        ModelArtifactNotFoundError
            If model artifact not found.
        """
        bucket_details = ObjectStorageDetails.from_path(bucket_uri)
        region = region or extract_region(self.auth)
        try:
            work_request_id = self.client.import_model_artifact(
                model_id=self.id,
                import_model_artifact_details=ImportModelArtifactDetails(
                    artifact_import_details=ArtifactImportDetailsObjectStorage(
                        namespace=bucket_details.namespace,
                        destination_bucket=bucket_details.bucket,
                        destination_object_name=bucket_details.filepath,
                        destination_region=region,
                    )
                ),
            ).headers["opc-work-request-id"]

            # Show progress of importing artifacts
            self._wait_for_work_request(
                work_request_id=work_request_id,
                num_steps=2,
            )
        except ServiceError as ex:
            if ex.status == 404:
                raise ModelArtifactNotFoundError()

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the artifact can be exported."
    )
    def export_model_artifact(self, bucket_uri: str, region: str = None):
        """Exports model artifact to the model catalog.
        Can be used for any model artifact. Requires to provide an Object Storage bucket,
        for transitional saving artifacts. For the small artifacts use `create_model_artifact` method.

        Parameters
        ----------
        bucket_uri: str
            The OCI Object Storage URI where model artifacts will be copied to.
            The `bucket_uri` is only necessary for downloading large artifacts which
            size is greater than 2GB.
            Example: `oci://<bucket_name>@<namespace>/prefix/`.
        region: (str, optional). Defaults to `None`.
            The destination Object Storage bucket region.
            By default the value will be extracted from the `OCI_REGION_METADATA` environment variables.

        Returns
        -------
        None
        """
        bucket_details = ObjectStorageDetails.from_path(bucket_uri)
        region = region or extract_region(self.auth)

        work_request_id = self.client.export_model_artifact(
            model_id=self.id,
            export_model_artifact_details=ExportModelArtifactDetails(
                artifact_export_details=ArtifactExportDetailsObjectStorage(
                    namespace=bucket_details.namespace,
                    source_bucket=bucket_details.bucket,
                    source_object_name=bucket_details.filepath,
                    source_region=region,
                )
            ),
        ).headers["opc-work-request-id"]

        # Show progress of exporting model artifacts
        self._wait_for_work_request(
            work_request_id=work_request_id,
            num_steps=3,
        )

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before it can be updated."
    )
    def update(self) -> "OCIDataScienceModel":
        """Updates datascience Model.

        Returns
        -------
        OCIDataScienceModel
            The `OCIDataScienceModel` instance (self).
        """
        return self.update_from_oci_model(
            self.client.update_model(
                self.id, self.to_oci_model(UpdateModelDetails)
            ).data
        )

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before it can be deleted."
    )
    def delete(
        self,
        delete_associated_model_deployment: Optional[bool] = False,
    ) -> "OCIDataScienceModel":
        """Deletes detascience Model.

        Parameters
        ----------
        delete_associated_model_deployment: (bool, optional). Defaults to `False`.
            Whether associated model deployments need to be deleted or not.

        Returns
        -------
        OCIDataScienceModel
            The `OCIDataScienceModel` instance (self).

        Raises
        ------
        ModelWithActiveDeploymentError
            If model has active deployments and `delete_associated_model_deployment` set to `False`.
        """
        active_deployments = tuple(
            item for item in self.model_deployment() if item.lifecycle_state == "ACTIVE"
        )

        if len(active_deployments) > 0:
            if not delete_associated_model_deployment:
                raise ModelWithActiveDeploymentError()

            logger.info(
                f"Deleting model deployments associated with the model `{self.id}`."
            )
            for oci_model_deployment in active_deployments:
                logger.info(
                    f"Deleting model deployment `{oci_model_deployment.identifier}`."
                )
                ModelDeployer(ds_client=self.client).delete(
                    model_deployment_id=oci_model_deployment.identifier
                )

        logger.info(f"Deleting model `{self.id}`.")
        self.client.delete_model(self.id)
        return self.sync()

    @check_for_model_id(
        msg="Model needs to be saved to the Model Catalog before the associated model deployments can be read."
    )
    def model_deployment(
        self,
        config: Optional[Dict] = None,
        tenant_id: Optional[str] = None,
        limit: Optional[int] = 500,
        page: Optional[str] = None,
        **kwargs: Dict,
    ) -> List:
        """
        Gets the list of model deployments by model ID across the compartments.

        Parameters
        ----------
        model_id: str
            The model ID.
        config: (Dict, optional). Defaults to `None`.
            Configuration keys and values as per SDK and Tool Configuration.
            The from_file() method can be used to load configuration from a file.
            Alternatively, a dict can be passed. You can validate_config the dict
            using validate_config(). Defaults to None.
        tenant_id: (str, optional). Defaults to `None`.
            The tenancy ID, which can be used to specify a different tenancy
            (for cross-tenancy authorization) when searching for resources in
            a different tenancy. Defaults to None.
        limit: (int, optional). Defaults to `None`.
            The maximum number of items to return. The value must be between
            1 and 1000. Defaults to 500.
        page: (str, optional). Defaults to `None`.
            The page at which to start retrieving results.

        Returns
        -------
            The list of model deployments associated with given model ID.
        """
        query = f"query datasciencemodeldeployment resources where ModelId='{self.id}'"
        return OCIResource.search(
            query,
            type=SEARCH_TYPE.STRUCTURED,
            config=config,
            tenant_id=tenant_id,
            limit=limit,
            page=page,
            **kwargs,
        )

    @classmethod
    def from_id(cls, ocid: str) -> "OCIDataScienceModel":
        """Gets model by OCID.

        Parameters
        ----------
        ocid: str
            The OCID of the datascience model.

        Returns
        -------
        OCIDataScienceModel
            An instance of `OCIDataScienceModel`.
        """
        if not ocid:
            raise ValueError("Model OCID not provided.")
        return super().from_ocid(ocid)

    def _wait_for_work_request(self, work_request_id: str, num_steps: int = 3) -> None:
        """Waits for the work request to be completed.

        Parameters
        ----------
        work_request_id: str
            Work Request OCID.
        num_steps: (int, optional). Defaults to 3.
            Number of steps for the progress indicator.

        Returns
        -------
        None
        """
        STOP_STATE = (
            WorkRequest.STATUS_SUCCEEDED,
            WorkRequest.STATUS_CANCELED,
            WorkRequest.STATUS_FAILED,
        )
        work_request_logs = []

        i = 0
        with utils.get_progress_bar(num_steps) as progress:
            while not work_request_logs or len(work_request_logs) < num_steps:
                time.sleep(_REQUEST_INTERVAL_IN_SEC)
                new_work_request_logs = []

                try:
                    work_request = self.client.get_work_request(work_request_id).data
                    work_request_logs = self.client.list_work_request_logs(
                        work_request_id
                    ).data
                except Exception as ex:
                    logger.warn(ex)

                new_work_request_logs = (
                    work_request_logs[i:] if work_request_logs else []
                )

                for wr_item in new_work_request_logs:
                    progress.update(wr_item.message)
                    i += 1

                if work_request and work_request.status in STOP_STATE:
                    if work_request.status != WorkRequest.STATUS_SUCCEEDED:
                        if new_work_request_logs:
                            raise Exception(new_work_request_logs[-1].message)
                        else:
                            raise Exception(
                                "Error occurred in attempt to perform the operation. "
                                "Check the service logs to get more details. "
                                f"{work_request}"
                            )
                    else:
                        break



#!/usr/bin/env python
# -*- coding: utf-8 -*--

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

from __future__ import print_function, absolute_import

import numpy as np
import pandas as pd

from collections import defaultdict, Counter
from itertools import combinations, product


def _chi_squared(count_matrix: np.ndarray, n_obs: int) -> float:
    """
    Computes Chi-squared when given a contingency table
    """
    row_sums = np.tile(np.sum(count_matrix, axis=1), (count_matrix.shape[1], 1)).T
    col_sums = np.tile(np.sum(count_matrix, axis=0), (count_matrix.shape[0], 1))
    return np.sum(
        np.square(count_matrix - row_sums * col_sums / n_obs)
        / (row_sums * col_sums / n_obs)
    )


def _cramers_v(cat1: np.ndarray, cat2: np.ndarray) -> float:
    """
    Calculates the cramers v of two numpy arrays.
    """
    keep_cat1 = ~pd.isnull(cat1)
    cat1_no_nan = cat1[keep_cat1]
    cat2_no_nan = cat2[keep_cat1]
    keep_cat2 = ~pd.isnull(cat2_no_nan)
    cat1_no_nan = cat1_no_nan[keep_cat2]
    cat2_no_nan = cat2_no_nan[keep_cat2]
    n = len(cat1_no_nan)
    if n == 1:
        return 0
    contingency_table, r, k = _count_occurrence(cat1_no_nan, cat2_no_nan)

    if r == 0:
        return 0.0000

    chi2 = _chi_squared(contingency_table, n)
    phi2 = chi2 / n

    phi2corr = max(0, phi2 - ((k - 1) * (r - 1)) / (n - 1))
    rcorr = r - (np.square(r - 1)) / (n - 1)
    kcorr = k - (np.square(k - 1)) / (n - 1)
    denominator = min((kcorr - 1), (rcorr - 1))
    if denominator == 0:
        return np.nan
    return np.sqrt(phi2corr / denominator)


def _list_to_dataframe(
    name_list: list, corr_list: list, normal_form: bool
) -> pd.DataFrame:
    corr_dict = defaultdict(dict)
    for idx, corr in zip(name_list, corr_list):
        row_name = idx[0]
        col_name = idx[1]
        corr_dict[row_name][col_name] = corr_dict[col_name][row_name] = round(corr, 4)
        corr_dict[row_name][row_name] = corr_dict[col_name][col_name] = 1.0000
    correlation_matrix = pd.DataFrame.from_dict(corr_dict).sort_index()
    correlation_matrix = correlation_matrix.loc[:, correlation_matrix.index]
    if normal_form:
        data = []
        for (col1, col2), corr in correlation_matrix.stack().iteritems():
            data.append([col1, col2, round(corr, 4)])
        return pd.DataFrame(data, columns=["Column 1", "Column 2", "Value"])
    else:
        return correlation_matrix


def _count_occurrence(cat1: np.ndarray, cat2: np.ndarray) -> (np.ndarray, int, int):
    """
    Calculates the contingency table of two arrays.
    """
    occurrence_cnt = Counter([(x, y) for x, y in zip(cat1, cat2)])
    nunique_cat1 = np.unique(cat1[~pd.isnull(cat1)])
    nunique_cat2 = np.unique(cat2[~pd.isnull(cat2)])
    r = len(nunique_cat1)
    k = len(nunique_cat2)
    contingency_table = np.zeros((r, k))
    for row, num1 in enumerate(nunique_cat1):
        for col, num2 in enumerate(nunique_cat2):
            contingency_table[row, col] = occurrence_cnt[(num1, num2)]

    return contingency_table, r, k


def _correlation_ratio(cat: np.ndarray, cts: np.ndarray):
    """
    Calculates the correlation of a pair of a category feature and a continuous feature
    using correlation ratio when input are two numpy arrays.
    """
    keep_cts = ~pd.isnull(cts)
    cat_no_nan = cat[keep_cts]
    cts_no_nan = cts[keep_cts]

    keep_cat = ~pd.isnull(cat_no_nan)
    cat_no_none = cat_no_nan[keep_cat]
    cts_no_none = cts_no_nan[keep_cat]

    unq_cat, tags, group_count = np.unique(
        list(cat_no_none), return_inverse=1, return_counts=1
    )
    group_mean = np.bincount(tags, cts_no_none) / group_count
    overall_mean = np.nanmean(cts_no_none)
    n = len(cts_no_none)

    dispersion_within = np.dot(group_count, np.square(group_mean - overall_mean))
    dispersion_population = cts_no_none.var() * n
    ratio = dispersion_within / dispersion_population

    return np.sqrt(ratio)


def cat_vs_cont(
    df: pd.DataFrame, categorical_columns, continuous_columns, normal_form: bool = True
) -> pd.DataFrame:
    """
    Calculates the correlation of all pairs of categorical features and continuous features.
    """
    numerical_categorical_pairs = list(product(categorical_columns, continuous_columns))
    corr_list = []
    for col in numerical_categorical_pairs:
        corr_list.append(
            _correlation_ratio(np.array(df[col[0]].values), np.array(df[col[1]].values))
        )
    correlation_matrix = _list_to_dataframe(
        numerical_categorical_pairs, corr_list, normal_form=normal_form
    )
    return correlation_matrix


def cat_vs_cat(df: pd.DataFrame, normal_form: bool = True) -> pd.DataFrame:
    """
    Calculates the correlation of all pairs of categorical features and categorical features.
    """
    categorical_columns = df.columns.to_list()
    categorical_pairs = list(combinations(categorical_columns, 2))
    corr_list = []
    for col in categorical_pairs:
        corr_list.append(
            _cramers_v(np.array(df[col[0]].values), np.array(df[col[1]].values))
        )
    correlation_matrix = _list_to_dataframe(
        categorical_pairs, corr_list, normal_form=normal_form
    )
    return correlation_matrix


def cont_vs_cont(df: pd.DataFrame, normal_form: bool = True) -> pd.DataFrame:
    """
    Calculates the Pearson correlation between two columns of the DataFrame.
    """
    if not normal_form:
        return df.corr(method="pearson")
    data = []
    for (col1, col2), corr in df.corr(method="pearson").stack().iteritems():
        data.append([col1, col2, round(corr, 4)])
    return pd.DataFrame(data, columns=["Column 1", "Column 2", "Value"])



#!/usr/bin/env python
# -*- coding: utf-8; -*-

# Copyright (c) 2021, 2022 Oracle and/or its affiliates.
# Licensed under the Universal Permissive License v 1.0 as shown at https://oss.oracle.com/licenses/upl/

"""Contains class wrapping OCI resource search service
"""
import oci.resource_search
from ads.common.oci_mixin import OCIClientMixin


class ResourceNotFoundError(Exception):
    """Exception when an OCI resource is not found or user does not have permission to access it.
    This could mean the resource does not exist, or
    there is not enough permission to find/access the resource.
    """


class SEARCH_TYPE(str):
    FREETEXT = "FreeText"
    STRUCTURED = "Structured"


class OCIResource(OCIClientMixin):
    """Contains helper methods for getting information from OCIResourceSearch service.

    Usage:
    Find the compartment ID of an OCI resource:
    >>> OCIResource.get_compartment_id("YOUR_OCID")
    Search for OCI resource matching free text (Similar to the search box in OCI console):
    >>> OCIResource.search("YOUR_FREE_TEXT")
    Search for OCI resource matching structured text:
    >>> OCIResource.search("STRUCTURED_TEXT", type="Structured")
    """

    @classmethod
    def init_client(cls, **kwargs) -> oci.resource_search.ResourceSearchClient:
        return cls._init_client(
            client=oci.resource_search.ResourceSearchClient, **kwargs
        )

    @classmethod
    def get_compartment_id(cls, ocid) -> str:
        """Gets the compartment OCID of an OCI resource, given the resource's OCID.

        Parameters
        ----------
        ocid : str
            OCID of a resource

        Returns
        -------
        str:
            Compartment OCID of the resource


        """
        results = set([r.compartment_id for r in cls.search(ocid)])
        if len(results) == 1:
            return list(results)[0]
        elif not results:
            raise ResourceNotFoundError(
                f"Resource not found for {ocid}. It may not exist or you may not have permission to list/access it."
            )
        else:
            raise ValueError(
                f"Unable determine compartment ID, multiple matches: {results}"
            )

    @classmethod
    def search(
        cls,
        query: str,
        type: str = "FreeText",
        config: dict = None,
        tenant_id: str = None,
        limit: int = 500,
        page: str = None,
        **kwargs,
    ) -> list:
        """Search OCI resource by free text.

        Parameters
        ----------
        query : str
            The content to search for.
        type :  str (optional)
            The type of SearchDetails, whether "FreeText" or "Structured".
            Defaults to "FreeText".
        config: dict (optional)
            Configuration keys and values as per SDK and Tool Configuration.
            The from_file() method can be used to load configuration from a file.
            Alternatively, a dict can be passed. You can validate_config the dict
            using validate_config(). Defaults to None.
        tenant_id: str (optional)
            The tenancy ID, which can be used to specify a different tenancy
            (for cross-tenancy authorization) when searching for resources in
            a different tenancy. Defaults to None.
        limit: int (optional)
            The maximum number of items to return. The value must be between
            1 and 1000. Defaults to 500.
        page: str (optional)
            The page at which to start retrieving results.

        Returns
        -------
        list
            A list of search results

        """
        if not config:
            client = cls.init_client()
        else:
            client = oci.resource_search.ResourceSearchClient(config=config)

        if type == "Structured":
            search_details = oci.resource_search.models.StructuredSearchDetails(
                type="Structured",
                query=query,
                matching_context_type=None,
            )
        elif type == "FreeText":
            search_details = oci.resource_search.models.FreeTextSearchDetails(
                text=query, type="FreeText"
            )
        else:
            raise ValueError(f"Invalid Type: {type}")

        return client.search_resources(
            search_details,
            limit=limit,
            page=page,
            tenant_id=tenant_id,
            **kwargs,
        ).data.items
